#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
"""
This module contains index statistics

It contains the following import shortcuts:
```python
from pygw.statistics.index import DifferingVisibilityCountStatistic
from pygw.statistics.index import DuplicateEntryCountStatistic
from pygw.statistics.index import FieldVisibilityCountStatistic
from pygw.statistics.index import IndexMetaDataSetStatistic
from pygw.statistics.index import MaxDuplicatesStatistic
from pygw.statistics.index import PartitionsStatistic
from pygw.statistics.index import RowRangeHistogramStatistic
```
"""

from .differing_visibility_count_statistic import DifferingVisibilityCountStatistic
from .duplicate_entry_count_statistic import DuplicateEntryCountStatistic
from .field_visibility_count_statistic import FieldVisibilityCountStatistic
from .index_meta_data_set_statistic import IndexMetaDataSetStatistic
from .max_duplicates_statistic import MaxDuplicatesStatistic
from .partitions_statistic import PartitionsStatistic
from .row_range_histogram_statistic import RowRangeHistogramStatistic



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from .base_query_builder import BaseQueryBuilder
from .aggregation_query import AggregationQuery
from ..base.type_conversions import StringArrayType


class AggregationQueryBuilder(BaseQueryBuilder):
    """
    A builder for creating aggregation queries. This class should not be used directly.  Instead, use one of the derived
    classes such as `pygw.query.vector.VectorAggregationQueryBuilder`.
    """

    def __init__(self, java_ref):
        super().__init__(java_ref)

    def count(self, *type_names):
        """
        This is a convenience method to set the count aggregation if no type names are given it is
        assumed to count every type.

        Args:
            type_names (str): The type names to count results.
        Returns:
            This query builder.
        """
        if type_names is None:
            self._java_ref.count()
        else:
            self._java_ref.count(StringArrayType().to_java(type_names))
        return self

    def aggregate(self, type_name, j_aggregation):
        """
        Provide the Java Aggregation function and the type name to apply the aggregation on.

        Args:
            type_name (str): The type name to aggregate.
            j_aggregation (Aggregation):  The Java aggregation function to
        Returns:
            This query builder.
        """

        return self._java_ref.aggregate(type_name, j_aggregation)

    def build(self):
        """
        Builds the configured aggregation query.

        Returns:
            The final constructed `pygw.query.AggregationQuery`.
        """
        return AggregationQuery(self._java_ref.build(), self._java_transformer)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from .geowave_object import GeoWaveObject
from .java_transformer import JavaTransformer
from ..config import java_pkg


class Envelope(GeoWaveObject):
    def __init__(self, min_x=0, min_y=0, max_x=0, max_y=0, java_ref=None):
        if java_ref is None:
            java_ref = java_pkg.org.locationtech.jts.geom.Envelope(
                float(min_x), float(max_x), float(min_y), float(max_y))
        super().__init__(java_ref)

    def get_min_x(self):
        return self._java_ref.getMinX()

    def get_max_x(self):
        return self._java_ref.getMaxX()

    def get_min_y(self):
        return self._java_ref.getMinY()

    def get_max_y(self):
        return self._java_ref.getMaxY()


class EnvelopeTransformer(JavaTransformer):
    """
    Transforms Java Envelopes into a pygw Envelope.
    """

    def transform(self, j_object):
        """
        Transform the given Java Envelope into a pygw Envelope.

        Args:
            j_object (Java Envelope): An Envelope Java object.
        Returns:
            A pygw implementation of Envelope.
        """
        return Envelope(java_ref=j_object)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from enum import Enum

from pygw.config import geowave_pkg


class QueryHintKey(Enum):
    """
    Keys for query hints.
    """

    MAX_RANGE_DECOMPOSITION = 0

    @classmethod
    def get_key(cls, key):
        """
        Gets the Java hint key from the given QueryHintKey.

        Args:
            key (pygw.query.query_hint_key.QueryHintKey): The enum value of QueryHintKey to get.
        Returns:
            The Java equivalent of the query hint key.
        """
        return {
            QueryHintKey.MAX_RANGE_DECOMPOSITION: geowave_pkg.core.store.util.DataStoreUtils.MAX_RANGE_DECOMPOSITION
        }.get(key)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base import GeoWaveObject
from .statistic_binning_strategy import StatisticBinningStrategy
from .statistic_type import DataTypeStatisticType, IndexStatisticType, FieldStatisticType
from .binning_strategy_mappings import map_binning_strategy
from ..base.java_transformer import NoOpTransformer


class Statistic(GeoWaveObject):
    """
    Base GeoWave statistic.
    """

    def __init__(self, java_ref, java_transformer=NoOpTransformer()):
        self.java_transformer = java_transformer
        super().__init__(java_ref)

    def get_statistic_type(self):
        """
        Get the statistic type associated with the statistic.

        Returns:
            The type of this statistic.
        """
        pass

    def get_description(self):
        """
        Gets a description of the statistic.

        Returns:
            A description of the statistic.
        """
        return self._java_ref.getDescription()

    def set_tag(self, tag):
        """
        Sets the tag of the statistic.

        Args:
            tag (str): The tag to use for the statistic
        """
        self._java_ref.setTag(tag)

    def get_tag(self):
        """
        Get the tag for the statistic.

        Returns:
            The tag for this statistic.
        """
        return self._java_ref.getTag()

    def set_internal(self):
        """
        Set the tag of this statistic to the default internal statistic tag.
        """
        self._java_ref.setInternal()

    def is_internal(self):
        """
        Checks if the statistic is an internal statistic.

        Returns:
            True if the statistic is internal.
        """
        return self._java_ref.isInternal()

    def set_binning_strategy(self, binning_strategy):
        """
        Sets the binning strategy of the statistic.

        Args:
            binning_strategy (StatisticBinningStrategy): The binning strategy to use for the statistic.
        """
        if not isinstance(binning_strategy, StatisticBinningStrategy):
            raise AttributeError('Expected an instance of StatisticBinningStrategy')
        self._java_ref.setBinningStrategy(binning_strategy.java_ref())

    def get_binning_strategy(self):
        """
        Gets the binning strategy used by the statistic.

        Returns:
            The binning strategy used by the statistic.
        """
        return map_binning_strategy(self._java_ref.getBinningStrategy())


class IndexStatistic(Statistic):

    def get_statistic_type(self):
        """
        Get the statistic type associated with the statistic.

        Returns:
            The type of this statistic.
        """
        return IndexStatisticType(self._java_ref.getStatisticType())

    def set_index_name(self, name):
        """
        Sets the index name of the statistic.

        Args:
            name (str): The index name to use for the statistic
        """
        self._java_ref.setIndexName(name)

    def get_index_name(self):
        """
        Get the index name associated with the statistic.

        Returns:
            The index name of this statistic.
        """
        return self._java_ref.getIndexName()


class DataTypeStatistic(Statistic):

    def get_statistic_type(self):
        """
        Get the statistic type associated with the statistic.

        Returns:
            The type of this statistic.
        """
        return DataTypeStatisticType(self._java_ref.getStatisticType())

    def set_type_name(self, name):
        """
        Sets the type name of the statistic.

        Args:
            name (str): The type name to use for the statistic
        """
        self._java_ref.setTypeName(name)

    def get_type_name(self):
        """
        Get the type name associated with the statistic.

        Returns:
            The type name of this statistic.
        """
        return self._java_ref.getTypeName()


class FieldStatistic(Statistic):

    def get_statistic_type(self):
        """
        Get the statistic type associated with the statistic.

        Returns:
            The type of this statistic.
        """
        return FieldStatisticType(self._java_ref.getStatisticType())

    def set_type_name(self, name):
        """
        Sets the type name of the statistic.

        Args:
            name (str): The type name to use for the statistic
        """
        self._java_ref.setTypeName(name)

    def get_type_name(self):
        """
        Get the type name associated with the statistic.

        Returns:
            The type name of this statistic.
        """
        return self._java_ref.getTypeName()

    def set_field_name(self, field_name):
        """
        Sets the field name of the statistic.

        Args:
            field_name (str): The field name to use for the statistic
        """
        self._java_ref.setFieldName(field_name)

    def get_field_name(self):
        """
        Get the field name associated with the statistic.

        Returns:
            The field name of this statistic.
        """
        return self._java_ref.getFieldName()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base import GeoWaveObject


class StatisticType(GeoWaveObject):
    """
    Base statistic type.
    """

    def __init__(self, java_ref):
        super().__init__(java_ref)

    def get_string(self):
        return self._java_ref.getString()


class IndexStatisticType(StatisticType):
    pass


class DataTypeStatisticType(StatisticType):
    pass


class FieldStatisticType(StatisticType):
    pass




#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
"""
This module contains statistic binning strategies

It contains the following import shortcuts:
```python
from pygw.statistics.binning_strategy import CompositeBinningStrategy
from pygw.statistics.binning_strategy import DataTypeBinningStrategy
from pygw.statistics.binning_strategy import FieldValueBinningStrategy
from pygw.statistics.binning_strategy import NumericRangeFieldValueBinningStrategy
from pygw.statistics.binning_strategy import SpatialFieldValueBinningStrategy
from pygw.statistics.binning_strategy import TimeRangeFieldValueBinningStrategy
from pygw.statistics.binning_strategy import PartitionBinningStrategy
```
"""

from .composite_binning_strategy import CompositeBinningStrategy
from .data_type_binning_strategy import DataTypeBinningStrategy
from .field_value_binning_strategy import FieldValueBinningStrategy
from .numeric_range_field_value_binning_strategy import NumericRangeFieldValueBinningStrategy
from .spatial_field_value_binning_strategy import SpatialFieldValueBinningStrategy
from .time_range_field_value_binning_strategy import TimeRangeFieldValueBinningStrategy
from .partition_binning_strategy import PartitionBinningStrategy



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from ..statistic_binning_strategy import StatisticBinningStrategy


class DataTypeBinningStrategy(StatisticBinningStrategy):
    """
    Statistic binning strategy that bins statistic values by data type name. This is generally only used for index
    statistics because data type and field statistics would all go under the same bin.
    """

    def __init__(self, java_ref=None):
        if java_ref is None:
            java_ref = geowave_pkg.core.store.statistics.binning.DataTypeBinningStrategy()
        super().__init__(java_ref)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.base import GeoWaveObject
from pygw.base.type_conversions import GeometryType
from . import AttributeDescriptor

from .simple_feature_type import SimpleFeatureType
from ..base.java_transformer import JavaTransformer


class SimpleFeature(GeoWaveObject):
    """
    A Simple (Vector) Feature.
    """

    def __init__(self, feature_type, java_ref):
        assert isinstance(feature_type, SimpleFeatureType)
        self._feature_type = feature_type
        super().__init__(java_ref)

    def get_id(self):
        """
        Returns:
            The ID of the feature.
        """
        return self._java_ref.getID()

    def get_type(self):
        """
        Returns:
            The `pygw.geotools.simple_feature_type.SimpleFeatureType` of the feature.
        """
        return self._feature_type

    def get_feature_type(self):
        """
        Identical to `get_type`.

        Returns:
            The `pygw.geotools.simple_feature_type.SimpleFeatureType` of the feature.
        """
        return self.get_type()

    def get_attributes(self):
        """
        Gets all of the attribute values of the feature in a single list.

        Returns:
            A list containing all attribute values of the feature.
        """
        j_values = self._java_ref.getAttributes()
        descriptors = self._feature_type.get_attribute_descriptors()
        values = []
        for i in range(len(j_values)):
            values.append(descriptors[i].from_java(j_values[i]))
        return values

    def get_attribute(self, attribute):
        """
        Gets a single attribute value by name or by index.

        Args:
            attribute (str or int): Name or index of the attribute to get.
        Returns:
            The value of the requested attribute.
        """
        j_value = self._java_ref.getAttribute(attribute)
        attr = self._feature_type.get_attribute(attribute)
        return attr.from_java(j_value)

    def get_attribute_count(self):
        """
        Returns:
            The number of attributes that the feature has.
        """
        return self._java_ref.getAttributeCount()

    def get_default_geometry(self):
        """
        Returns:
            The default geometry of the feature.
        """
        j_geom = self._java_ref.getDefaultGeometry()
        return GeometryType().from_java(j_geom)

    def to_dict(self, id_column="id"):
        """
        Convert this feature to a dictionary, including the feature ID.

        Args:
            id_column (str): The key for the feature ID. Default is "id".
        Returns:
            The attributes and feature ID of this feature in a dictionary.
        """
        j_values = self._java_ref.getAttributes()
        descriptors = self._feature_type.get_attribute_descriptors()
        feature_dict = {
            id_column: self._java_ref.getID()
        }
        for i in range(len(j_values)):
            feature_dict[descriptors[i].descriptor] = descriptors[i].from_java(j_values[i])
        return feature_dict


class SimpleFeatureTransformer(JavaTransformer):
    """
    Transforms Java SimpleFeatures into pgyw.geotools.SimpleFeatures.  In order to accomplish this, the pygw variant of
    the SimpleFeatureType has to be constructed from the feature.  In order to avoid doing this multiple times, there
    is a feature type cache that the transform function can pull from.
    """

    def __init__(self):
        self._feature_type_cache = {}

    def transform(self, j_object):
        """
        Transform the given Java SimpleFeature into a pygw.geotools.SimpleFeature.

        Args:
            j_object (Java SimpleFeature): A geotools SimpleFeature Java object.
        Returns:
            A `pygw.geotools.simple_feature.SimpleFeature`.
        """
        j_sft = j_object.getFeatureType()
        type_name = j_sft.getTypeName()
        if type_name in self._feature_type_cache:
            sft = self._feature_type_cache[type_name]
        else:
            j_attrs = j_sft.getAttributeDescriptors().iterator()
            descriptors = []
            while j_attrs.hasNext():
                descriptors.append(AttributeDescriptor.from_java_attribute_descriptor(j_attrs.next()))
            sft = SimpleFeatureType(j_sft, descriptors)
            self._feature_type_cache[type_name] = sft
        return SimpleFeature(sft, j_object)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.config import geowave_pkg
from pygw.store import DataStoreOptions


class AccumuloOptions(DataStoreOptions):
    """
    Accumulo data store options.
    """

    def __init__(self):
        super().__init__(geowave_pkg.datastore.accumulo.config.AccumuloRequiredOptions())

    def set_zookeeper(self, zookeeper):
        """
        Sets the list of Zookeper servers that the Accumulo instance uses as a comma-separated
        string.

        Args:
            zookeeper (str): A comma-separated list of Zookeeper servers.
        """
        self._java_ref.setZookeeper(zookeeper)

    def get_zookeeper(self):
        """
        Returns:
            A comma-separated list of Zookeper servers.
        """
        return self._java_ref.getZookeeper()

    def set_instance(self, instance):
        """
        Sets the Accumulo instance ID to use for the data store.

        Args:
            instance (str): The Accumulo instance ID to use.
        """
        self._java_ref.setInstance(instance)

    def get_instance(self):
        """
        Returns:
            The Accumulo instance ID.
        """
        return self._java_ref.getInstance()

    def set_user(self, user):
        """
        Sets the Accumulo user ID.

        Args:
            user (str): The Accumulo user ID.
        """
        self._java_ref.setUser(user)

    def get_user(self):
        """
        Returns:
            The Accumulo user ID.
        """
        return self._java_ref.getUser()

    def set_password(self, password):
        """
        Sets the Accumulo password.

        Args:
            password (str): The Accumulo password.
        """
        self._java_ref.setPassword(password)

    def get_password(self):
        """
        Returns:
            The Accumulo password.
        """
        return self._java_ref.getPassword()

    def set_use_locality_groups(self, use_locality_groups):
        """
        Sets whether or not to use locality groups.

        Args:
            use_locality_groups (bool): Whether or not to use locality groups.
        """
        self._base_options.setUseLocalityGroups(use_locality_groups)

    def is_use_locality_groups(self):
        """
        Returns:
            True if locality groups are enabled, False otherwise.
        """
        return self._base_options.isUseLocalityGroups()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from .statistic import Statistic
from .data_type.count_statistic import CountStatistic
from .field.bloom_filter_statistic import BloomFilterStatistic
from .field.bounding_box_statistic import BoundingBoxStatistic
from .field.count_min_sketch_statistic import CountMinSketchStatistic
from .field.fixed_bin_numeric_histogram_statistic import FixedBinNumericHistogramStatistic
from .field.hyper_log_log_statistic import HyperLogLogStatistic
from .field.numeric_histogram_statistic import NumericHistogramStatistic
from .field.numeric_mean_statistic import NumericMeanStatistic
from .field.numeric_range_statistic import NumericRangeStatistic
from .field.numeric_stats_statistic import NumericStatsStatistic
from .field.time_range_statistic import TimeRangeStatistic
from .index.differing_visibility_count_statistic import DifferingVisibilityCountStatistic
from .index.duplicate_entry_count_statistic import DuplicateEntryCountStatistic
from .index.field_visibility_count_statistic import FieldVisibilityCountStatistic
from .index.index_meta_data_set_statistic import IndexMetaDataSetStatistic
from .index.max_duplicates_statistic import MaxDuplicatesStatistic
from .index.partitions_statistic import PartitionsStatistic
from .index.row_range_histogram_statistic import RowRangeHistogramStatistic


__statistic_mappings = {
    CountStatistic.STATS_TYPE.get_string(): CountStatistic,
    BloomFilterStatistic.STATS_TYPE.get_string(): BloomFilterStatistic,
    BoundingBoxStatistic.STATS_TYPE.get_string(): BoundingBoxStatistic,
    CountMinSketchStatistic.STATS_TYPE.get_string(): CountMinSketchStatistic,
    FixedBinNumericHistogramStatistic.STATS_TYPE.get_string(): FixedBinNumericHistogramStatistic,
    HyperLogLogStatistic.STATS_TYPE.get_string(): HyperLogLogStatistic,
    NumericHistogramStatistic.STATS_TYPE.get_string(): NumericHistogramStatistic,
    NumericMeanStatistic.STATS_TYPE.get_string(): NumericMeanStatistic,
    NumericRangeStatistic.STATS_TYPE.get_string(): NumericRangeStatistic,
    NumericStatsStatistic.STATS_TYPE.get_string(): NumericStatsStatistic,
    TimeRangeStatistic.STATS_TYPE.get_string(): TimeRangeStatistic,
    DifferingVisibilityCountStatistic.STATS_TYPE.get_string(): DifferingVisibilityCountStatistic,
    DuplicateEntryCountStatistic.STATS_TYPE.get_string(): DuplicateEntryCountStatistic,
    FieldVisibilityCountStatistic.STATS_TYPE.get_string(): FieldVisibilityCountStatistic,
    IndexMetaDataSetStatistic.STATS_TYPE.get_string(): IndexMetaDataSetStatistic,
    MaxDuplicatesStatistic.STATS_TYPE.get_string(): MaxDuplicatesStatistic,
    PartitionsStatistic.STATS_TYPE.get_string(): PartitionsStatistic,
    RowRangeHistogramStatistic.STATS_TYPE.get_string(): RowRangeHistogramStatistic,
}


def map_statistic(j_statistic):
    if j_statistic is None:
        return None
    stat_type = j_statistic.getStatisticType().getString()
    if stat_type in __statistic_mappings:
        return __statistic_mappings[stat_type](java_ref=j_statistic)
    return Statistic(j_statistic)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base import GeoWaveObject
from ..config import geowave_pkg


class StatisticBinningStrategy(GeoWaveObject):
    """
    Base statistic binning strategy class.
    """

    def __init__(self, java_ref):
        super().__init__(java_ref)

    def get_strategy_name(self):
        """
        Gets the name of the binning strategy.

        Returns:
            The name of the binning strategy.
        """
        return self._java_ref.getStrategyName()

    def get_description(self):
        """
        Gets a description of the binning strategy.

        Returns:
            A description of the binning strategy.
        """
        return self._java_ref.getDescription()

    def bin_to_string(self, stat_bin):
        """
        Convert a bin to a readable string.

        Args:
            stat_bin (bytes): The bin to convert to string.
        Returns:
            A string that represents the bin.
        """
        j_byte_array = geowave_pkg.core.index.ByteArray(stat_bin)
        return self._java_ref.binToString(j_byte_array)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base import GeoWaveObject
from pygw.config import geowave_pkg
from ..statistic import FieldStatistic
from ..statistic_type import FieldStatisticType
from ...base.java_transformer import JavaTransformer


class BloomFilterStatistic(FieldStatistic):
    """
    Applies a bloom filter to field values useful for quickly determining set membership. False positives are possible
    but false negatives are not possible. In other words, a value can be determined to be possibly in the set or
    definitely not in the set.
    """
    STATS_TYPE = FieldStatisticType(geowave_pkg.core.store.statistics.field.BloomFilterStatistic.STATS_TYPE)

    def __init__(self, type_name=None, field_name=None, java_ref=None):
        if java_ref is None:
            if type_name is None and field_name is None:
                java_ref = geowave_pkg.core.store.statistics.field.BloomFilterStatistic()
            else:
                java_ref = geowave_pkg.core.store.statistics.field.BloomFilterStatistic(type_name, field_name)
        super().__init__(java_ref, BloomFilterTransformer())

    def set_expected_insertions(self, expected_insertions):
        self._java_ref.setExpectedInsertions(int(expected_insertions))

    def get_expected_insertions(self):
        return self._java_ref.getExpectedInsertions()

    def set_desired_false_positive_probability(self, probability):
        self._java_ref.setDesiredFalsePositiveProbability(float(probability))

    def get_desired_false_positive_probability(self):
        return self._java_ref.getDesiredFalsePositiveProbability()


class BloomFilterTransformer(JavaTransformer):
    def transform(self, j_object):
        return BloomFilter(j_object)


class BloomFilter(GeoWaveObject):
    def might_contain(self, value):
        return self._java_ref.mightContain(value)

    def expected_fpp(self):
        return self._java_ref.expectedFpp()

    def approximate_element_count(self):
        return self._java_ref.approximateElementCount()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base import GeoWaveObject
from pygw.base.type_conversions import StringArrayType
from pygw.config import geowave_pkg
from pygw.query.statistics.statistic_query import StatisticQuery
from pygw.statistics.bin_constraints import BinConstraints
from pygw.statistics.statistic_type import StatisticType, IndexStatisticType, DataTypeStatisticType


class StatisticQueryBuilder(GeoWaveObject):
    """
    A builder for creating statistics queries. This class should not be constructed directly, instead use one of the
    static methods to create an appropriate builder.
    """

    def __init__(self, java_ref, result_transformer):
        self._result_transformer = result_transformer
        super().__init__(java_ref)

    def tag(self, tag):
        """
        Sets the tag to query for.

        Args:
            tag (str): The tag to query for.
        Returns:
            This statistic query builder.
        """
        self._java_ref.tag(tag)
        return self

    def internal(self):
        """
        When set, only internal statistics will be queried.

        Returns:
            This statistic query builder.
        """
        self._java_ref.internal()
        return self

    def add_authorization(self, authorization):
        """
        Adds an authorization to the query.

        Args:
            authorization (str): The authorization to add.
        Returns:
            This statistic query builder.
        """
        self._java_ref.addAuthorization(authorization)
        return self

    def authorizations(self, authorizations):
        """
        Sets the set of authorizations to use for the query.

        Args:
            authorizations (array of str): The authorizations to use for the query.
        Returns:
            This statistic query builder.
        """
        self._java_ref.authorizations(StringArrayType().to_java(authorizations))
        return self

    def bin_constraints(self, bin_constraints):
        """
        Sets the constraints to use for the statistic query.  Only bins that match the given constraints will be
        returned.

        Args:
            bin_constraints (BinConstraints): The constraints to constrain the query by.
        Returns:
            This statistic query builder.
        """
        if not isinstance(bin_constraints, BinConstraints):
            raise AttributeError('Must be a BinConstraints instance.')
        self._java_ref.binConstraints(bin_constraints.java_ref())
        return self

    def build(self):
        """
        Build the statistic query.

        Returns:
            This constructed statistic query.
        """
        return StatisticQuery(self._java_ref.build(), self._result_transformer)

    @staticmethod
    def new_builder(statistic_type):
        """
        Create a statistic query builder for the given statistic type.

        Args:
            statistic_type (StatisticType): The statistic type for the query builder.
        Returns:
            A statistic query builder.
        """
        if not isinstance(statistic_type, StatisticType):
            raise AttributeError('Must be a StatisticType instance.')

        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.newBuilder(statistic_type.java_ref())
        if isinstance(statistic_type, IndexStatisticType):
            return IndexStatisticQueryBuilder(statistic_type, j_builder)
        if isinstance(statistic_type, DataTypeStatisticType):
            return DataTypeStatisticQueryBuilder(statistic_type, j_builder)
        return FieldStatisticQueryBuilder(statistic_type, j_builder)

    @staticmethod
    def differing_visibility_count():
        """
        Create a statistic query builder for a differing visibility count statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.differingVisibilityCount()
        return IndexStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def duplicate_entry_count():
        """
        Create a statistic query builder for a duplicate entry count statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.duplicateEntryCount()
        return IndexStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def field_visibility_count():
        """
        Create a statistic query builder for a field visibility count statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.fieldVisibilityCount()
        return IndexStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def index_meta_data_set():
        """
        Create a statistic query builder for an index meta data set statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.indexMetaDataSet()
        return IndexStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def max_duplicates():
        """
        Create a statistic query builder for a max duplicates statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.maxDuplicates()
        return IndexStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def partitions():
        """
        Create a statistic query builder for a partitions statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.partitions()
        return IndexStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def row_range_histogram():
        """
        Create a statistic query builder for a row range histogram statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.rowRangeHistogram()
        return IndexStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def count():
        """
        Create a statistic query builder for a count statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.count()
        return DataTypeStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def bbox():
        """
        Create a statistic query builder for a bounding box statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.geotime.store.statistics.SpatialTemporalStatisticQueryBuilder.bbox()
        return FieldStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def bloom_filter():
        """
        Create a statistic query builder for a bloom filter statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.bloomFilter()
        return FieldStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def time_range():
        """
        Create a statistic query builder for a time range statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.geotime.store.statistics.SpatialTemporalStatisticQueryBuilder.timeRange()
        return FieldStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def count_min_sketch():
        """
        Create a statistic query builder for a count min sketch statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.countMinSketch()
        return FieldStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def fixed_bin_numeric_histogram():
        """
        Create a statistic query builder for a fixed bin numeric histogram statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.fixedBinNumericHistogram()
        return FieldStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def hyper_log_log():
        """
        Create a statistic query builder for a hyper log log statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.hyperLogLog()
        return FieldStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def numeric_histogram():
        """
        Create a statistic query builder for a numeric histogram statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.numericHistogram()
        return FieldStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def numeric_mean():
        """
        Create a statistic query builder for a numeric mean statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.numericMean()
        return FieldStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def numeric_range():
        """
        Create a statistic query builder for a numeric range statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.numericRange()
        return FieldStatisticQueryBuilder(java_ref=j_builder)

    @staticmethod
    def numeric_stats():
        """
        Create a statistic query builder for a numeric stats statistic.

        Returns:
            A statistic query builder.
        """
        j_builder = geowave_pkg.core.store.api.StatisticQueryBuilder.numericStats()
        return FieldStatisticQueryBuilder(java_ref=j_builder)


class IndexStatisticQueryBuilder(StatisticQueryBuilder):
    """
    A builder for index statistic queries.
    """

    def __init__(self, statistic_type=None, java_ref=None):
        if java_ref is None:
            j_qbuilder = geowave_pkg.core.statistics.query.IndexStatisticQueryBuilder(statistic_type.java_ref())
        else:
            j_qbuilder = java_ref
        super().__init__(j_qbuilder, None)

    def index_name(self, index_name):
        """
        Set the index name to constrain the query by.

        Args:
            index_name (str): The index name to query.
        Returns:
            This statistic query builder.
        """
        self._java_ref.indexName(index_name)
        return self


class DataTypeStatisticQueryBuilder(StatisticQueryBuilder):
    """
    A builder for data type statistic queries.
    """

    def __init__(self, statistic_type=None, java_ref=None):
        if java_ref is None:
            j_qbuilder = geowave_pkg.core.statistics.query.DataTypeStatisticQueryBuilder(statistic_type.java_ref())
        else:
            j_qbuilder = java_ref
        super().__init__(j_qbuilder, None)

    def type_name(self, type_name):
        """
        Set the type name to constrain the query by.

        Args:
            type_name (str): The type name to query.
        Returns:
            This statistic query builder.
        """
        self._java_ref.typeName(type_name)
        return self


class FieldStatisticQueryBuilder(StatisticQueryBuilder):
    """
    A builder for field statistic queries.
    """

    def __init__(self, statistic_type=None, java_ref=None):
        if java_ref is None:
            j_qbuilder = geowave_pkg.core.statistics.query.FieldStatisticQueryBuilder(statistic_type.java_ref())
        else:
            j_qbuilder = java_ref
        super().__init__(j_qbuilder, None)

    def type_name(self, type_name):
        """
        Set the type name to constrain the query by.

        Args:
            type_name (str): The type name to query.
        Returns:
            This statistic query builder.
        """
        self._java_ref.typeName(type_name)
        return self

    def field_name(self, field_name):
        """
        Set the field name to constrain the query by.

        Args:
            field_name (str): The field name to query.
        Returns:
            This statistic query builder.
        """
        self._java_ref.fieldName(field_name)
        return self



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base import GeoWaveObject
from pygw.base.type_conversions import PrimitiveByteArrayType, GeometryType
from pygw.config import geowave_pkg, java_gateway
from shapely.geometry.base import BaseGeometry


class BinConstraints(GeoWaveObject):
    """
    Bin constraints for statistic queries.
    """

    def __init__(self, java_ref):
        super().__init__(java_ref)

    @staticmethod
    def _convert_byte_arrays(*byte_arrays):
        n = len(byte_arrays)
        j_arr = java_gateway.new_array(geowave_pkg.core.index.ByteArray, n)
        byte_array_type = PrimitiveByteArrayType()
        for idx, byte_array in enumerate(byte_arrays):
            j_arr[idx] = geowave_pkg.core.index.ByteArray(byte_array_type.to_java(byte_array))

        return j_arr

    @staticmethod
    def all_bins():
        """
        Unconstrained, a query will return all of the bins.

        Returns:
            BinConstraints that represent all bins.
        """
        j_constraints = geowave_pkg.core.store.api.BinConstraints.allBins()
        return BinConstraints(j_constraints)

    @staticmethod
    def of(*exact_match_bins):
        """
        Sets the bins of the query explicitly. If a queried statistic uses a binning strategy, only values contained in
        one of the given bins will be return.

        Args:
            exact_match_bins (bytes): The bins to match.
        Returns:
            A bin constraint representing exact matches of the provided bins.
        """
        j_constraints = geowave_pkg.core.store.api.BinConstraints.of(
            BinConstraints._convert_byte_arrays(*exact_match_bins))
        return BinConstraints(j_constraints)

    @staticmethod
    def of_prefix(*prefix_bins):
        """
        Sets the bins of the query by prefix. If a queried statistic uses a binning strategy, only values matching the
        bin prefix will be returned.

        Args:
            prefix_bins (bytes): The prefixes used to match the bins.
        Returns:
            A bin constraint representing the set of bin prefixes.
        """
        j_constraints = geowave_pkg.core.store.api.BinConstraints.ofPrefix(
            BinConstraints._convert_byte_arrays(*prefix_bins))
        return BinConstraints(j_constraints)

    @staticmethod
    def of_object(binning_strategy_constraint):
        """
        Sets the bins of the query using an object type that is supported by the binning strategy. The result will be
        constrained to only statistics that use binning strategies that support this type of constraint and the
        resulting bins will be constrained according to that strategy's usage of this object. For example, spatial
        binning strategies may use spatial Envelope as constraints, or another example might be a numeric field binning
        strategy using Range as constraints. If a queried statistic uses a binning strategy, only values
        contained in one of the given bins will be return.

        Args:
            binning_strategy_constraint (any): An object of any type supported by the binning strategy. It will be
            interpreted as appropriate by the binning strategy and binning strategies that do not support this object
            type will not return any results.
        Returns:
            Bin constraints representing the provided object.
        """
        if isinstance(binning_strategy_constraint, GeoWaveObject):
            binning_strategy_constraint = binning_strategy_constraint.java_ref()
        elif isinstance(binning_strategy_constraint, BaseGeometry):
            binning_strategy_constraint = GeometryType().to_java(binning_strategy_constraint)
        j_constraints = geowave_pkg.core.store.api.BinConstraints.ofObject(binning_strategy_constraint)
        return BinConstraints(j_constraints)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.config import geowave_pkg
from pygw.base import DataTypeAdapter

from .simple_feature_type import SimpleFeatureType


class FeatureDataAdapter(DataTypeAdapter):
    """
    Data adapter for storing and retrieving SimpleFeatures from a GeoWave data store.
    """

    def __init__(self, feature_type):
        """
        Constructs a feature data adapter using the given feature type.

        Args:
            feature_type (pygw.geotools.simple_feature_type.SimpleFeatureType): The feature type of the data.
        """
        self.feature_type = feature_type
        assert isinstance(feature_type, SimpleFeatureType)
        j_feat_type = feature_type._java_ref
        j_feat_adapter = geowave_pkg.adapter.vector.FeatureDataAdapter(j_feat_type)
        super().__init__(j_feat_adapter)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.config import geowave_pkg

from ...geotools.simple_feature import SimpleFeatureTransformer
from ..query_builder import QueryBuilder
from .vector_query_constraints_factory import VectorQueryConstraintsFactory


class VectorQueryBuilder(QueryBuilder):
    """
    A builder for creating queries for vector data.
    """

    def __init__(self):
        j_vector_qbuilder = geowave_pkg.core.geotime.store.query.api.VectorQueryBuilder.newBuilder()
        super().__init__(j_vector_qbuilder, SimpleFeatureTransformer())

    def constraints_factory(self):
        """
        Creates a constraints factory for vector queries.  The vector query constraint factory
        provides additional constraints specific to vector data.

        Returns:
            A `pygw.query.vector.vector_query_constraints_factory.VectorQueryConstraintsFactory`.
        """
        return VectorQueryConstraintsFactory(self._java_ref.constraintsFactory())



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.config import geowave_pkg
from pygw.store import DataStoreOptions


class HBaseOptions(DataStoreOptions):
    """
    HBase data store options.
    """

    def __init__(self):
        super().__init__(geowave_pkg.datastore.hbase.config.HBaseRequiredOptions())

    def set_zookeeper(self, zookeeper):
        """
        Sets the list of Zookeper servers that the HBase instance uses as a comma-separated
        string.

        Args:
            zookeeper (str): A comma-separated list of Zookeeper servers.
        """
        self._java_ref.setZookeeper(zookeeper)

    def get_zookeeper(self):
        """
        Returns:
            A comma-separated list of Zookeper servers.
        """
        return self._java_ref.getZookeeper()

    def set_scan_cache_size(self, scan_cache_size):
        """
        Sets the scan cache size of the HBase instance.

        Args:
            scan_cache_size (int): The scan cache size to use.
        """
        self._base_options.setScanCacheSize(scan_cache_size)

    def get_scan_cache_size(self):
        """
        Returns:
            The scan cache size of the HBase instance.
        """
        return self._base_options.getScanCacheSize()

    def set_verify_coprocessors(self, verify_coprocessors):
        """
        Sets whether or not to verify coprocessors when performing operations.

        Args:
            verify_coprocessors (bool): Whether or not to verify coprocessors.
        """
        self._base_options.setVerifyCoprocessors(verify_coprocessors)

    def is_verify_coprocessors(self):
        """
        Returns:
            True if coprocessors will be verified, False otherwise.
        """
        return self._base_options.isVerifyCoprocessors()

    def set_coprocessor_jar(self, coprocessor_jar):
        """
        Sets the path (HDFS URL) to the jar containing coprocessor classes.

        Args:
            coprocessor_jar (str): The path to the coprocessor jar.
        """
        self._base_options.setCoprocessorJar(coprocessor_jar)

    def get_coprocessor_jar(self):
        """
        Returns:
            The HDFS URL of the coprocessor jar.
        """
        return self._base_options.getCoprocessorJar()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from datetime import datetime, timezone

from .geowave_object import GeoWaveObject
from .java_transformer import JavaTransformer
from ..config import java_pkg


class Interval(GeoWaveObject):
    def __init__(self, start=datetime.now(), end=datetime.now(), java_ref=None):
        if java_ref is None:
            j_start = java_pkg.java.time.Instant.ofEpochSecond(int(start.replace(tzinfo=timezone.utc).timestamp()))
            j_end = java_pkg.java.time.Instant.ofEpochSecond(int(end.replace(tzinfo=timezone.utc).timestamp()))
            java_ref = java_pkg.org.threeten.extra.Interval.of(j_start, j_end)
        super().__init__(java_ref)

    def get_start(self):
        return datetime.utcfromtimestamp(self._java_ref.getStart().getEpochSecond())

    def get_end(self):
        return datetime.utcfromtimestamp(self._java_ref.getEnd().getEpochSecond())


class IntervalTransformer(JavaTransformer):
    """
    Transforms Java Intervals into a pygw Interval.
    """

    def transform(self, j_object):
        """
        Transform the given Java Interval into a pygw Interval.

        Args:
            j_object (Java Interval): An Interval Java object.
        Returns:
            A pygw implementation of Interval.
        """
        return Interval(java_ref=j_object)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
"""
This module includes common classes that are used by other modules.  This includes the base `GeoWaveObject` class that
serves as a python wrapper for a java reference.  It also includes a `type_conversions` submodule that can be used to
convert Python types to Java types that are commonly used in GeoWave.

It contains the following import shortcuts:
```python
from pygw.base import GeoWaveObject
from pygw.base import CloseableIterator
from pygw.base import DataTypeAdapter
from pygw.base import Writer
from pygw.base import Envelope
from pygw.base import Interval
from pygw.base import Range
```
"""

from .geowave_object import GeoWaveObject
from .closeable_iterator import CloseableIterator
from .data_type_adapter import DataTypeAdapter
from .writer import Writer
from .envelope import Envelope
from .interval import Interval
from .range import Range



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
"""
This module contains classes that are used in constructing queries and their constraints.

It contains the following import shortcuts:
```python
from pygw.query import Query
from pygw.query import QueryConstraints
from pygw.query import QueryConstraintsFactory
from pygw.query import QueryHintKey
from pygw.query import AggregationQuery
from pygw.query import VectorQueryBuilder
from pygw.query import VectorAggregationQueryBuilder
from pygw.query import FilterFactory
```
"""

from .query import Query
from .query_constraints import QueryConstraints
from .query_constraints_factory import QueryConstraintsFactory
from .query_hint_key import QueryHintKey
from .aggregation_query import AggregationQuery
from .vector import VectorAggregationQueryBuilder
from .vector import VectorQueryBuilder
from .vector import FilterFactory



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from ..query_constraints_factory import QueryConstraintsFactory
from ..query_constraints import QueryConstraints
from .spatial_temporal_constraints_builder import SpatialTemporalConstraintsBuilder


class VectorQueryConstraintsFactory(QueryConstraintsFactory):
    """
    A query constraints factory with additional methods for creating spatial and/or
    temporal constraints for vector data.  Do not construct this class manually, instead,
    get the constraints factory by using the `constraints_factory()` method of the
    query builder.
    """

    def spatial_temporal_constraints(self):
        """
        Creates a spatial temporal constraints builder that can be used to construct
        spatial and/or temporal constraints.

        Returns:
            A new `pygw.query.vector.spatial_temporal_constraints_builder.SpatialTemporalConstraintsBuilder`.
        """
        return SpatialTemporalConstraintsBuilder(self._java_ref.spatialTemporalConstraints())

    def filter_constraints(self, filter_constraint):
        """
        Constrain a query using a filter created by pygw.query.FilterFactory.

        Args:
            filter_constraint (filter): The filter to constrain the query by.
        Returns:
            A `pygw.query.query_constraints.QueryConstraints` with the given filter.
        """
        return QueryConstraints(self._java_ref.filterConstraints(filter_constraint))

    def cql_constraints(self, cql_expression):
        """
        Constrain a query using a CQL expression.

        Args:
            cql_expression (str): The CQL expression to constrain the query by.
        Returns:
            A `pygw.query.query_constraints.QueryConstraints` with the given CQL expression.
        """
        return QueryConstraints(self._java_ref.cqlConstraints(cql_expression))



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from ..statistic import FieldStatistic
from ..statistic_type import FieldStatisticType
from ...base import GeoWaveObject
from ...base.java_transformer import JavaTransformer


class CountMinSketchStatistic(FieldStatistic):
    """
    Maintains an estimate of how may of each attribute value occurs in a set of data.
    """
    STATS_TYPE = FieldStatisticType(geowave_pkg.core.store.statistics.field.CountMinSketchStatistic.STATS_TYPE)

    def __init__(
            self,
            type_name=None,
            field_name=None,
            error_factor=0.001,
            probability_of_correctness=0.98,
            java_ref=None):
        if java_ref is None:
            if type_name is None and field_name is None:
                java_ref = geowave_pkg.core.store.statistics.field.CountMinSketchStatistic()
            else:
                java_ref = geowave_pkg.core.store.statistics.field.CountMinSketchStatistic(
                    type_name, field_name, float(error_factor), float(probability_of_correctness))
        super().__init__(java_ref, CountMinSketchTransformer())

    def set_error_factor(self, error_factor):
        self._java_ref.setErrorFactor(float(error_factor))

    def get_error_factor(self):
        return self._java_ref.getErrorFactor()

    def set_probability_of_correctness(self, probability_of_correctness):
        self._java_ref.setProbabilityOfCorrectness(float(probability_of_correctness))

    def get_probability_of_correctness(self):
        return self._java_ref.getProbabilityOfCorrectness()


class CountMinSketchTransformer(JavaTransformer):
    def transform(self, j_object):
        return CountMinSketch(j_object)


class CountMinSketch(GeoWaveObject):
    def get_relative_error(self):
        return self._java_ref.getRelativeError()

    def get_confidence(self):
        return self._java_ref.getConfidence()

    def estimate_count(self, item):
        return self._java_ref.estimateCount(item)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from decimal import Decimal
from datetime import date, timezone
from datetime import datetime

from py4j.java_gateway import JavaClass
from shapely.geometry.base import BaseGeometry
from shapely.geometry import Point
from shapely.geometry import LineString
from shapely.geometry import Polygon
from shapely.geometry import MultiPoint
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPolygon
from shapely.geometry import GeometryCollection
from shapely.wkb import dumps
from shapely.wkb import loads

from pygw.config import java_pkg
from pygw.config import java_gateway

_wkb_reader = java_pkg.org.locationtech.jts.io.WKBReader()
_wkb_writer = java_pkg.org.locationtech.jts.io.WKBWriter()


def _type_to_string(py_type):
    if isinstance(py_type, tuple):
        return ", ".join(t.__name__ for t in py_type)
    else:
        return py_type.__name__


class AttributeType:
    """
    Base class for attributes that can be converted to and from Java variants.
    """

    def __init__(self, binding, py_type):
        self.binding = binding
        self._py_type = py_type

    def to_java(self, value):
        """
        Convert a Python variable into its Java counterpart.

        Args:
            value (any): The Python variable to convert.
        Returns:
            The Java counterpart of the Python variable.
        """
        if value is None:
            return value
        if isinstance(value, self._py_type):
            return self._to_java(value)
        else:
            self._value_error(value, _type_to_string(self._py_type))

    def _to_java(self, value):
        return value

    def from_java(self, value):
        """
        Convert a Java variable into its Python counterpart.

        Args:
            value (any): The Java variable to convert.
        Returns:
            The Python counterpart of the Java variable.
        """
        if value is None:
            return value
        else:
            return self._from_java(value)

    def _from_java(self, value):
        return value

    def _value_error(self, value, expected):
        raise ValueError("Value[%s] should be of type %s." % (str(value), expected))


class ArrayAttributeType(AttributeType):
    """
    Base class for attributes that represent an array of values.
    """

    def __init__(self, subtype, py_type=(list, tuple)):
        self.subtype = subtype
        self._j_class = JavaClass(self.subtype.binding, java_gateway._gateway_client)
        super().__init__("[L%s;" % self.subtype.binding, py_type)

    def _to_java(self, value):
        j_arr = java_gateway.new_array(self._j_class, len(value))
        for i in range(len(value)):
            if value[i] is None:
                continue
            j_arr[i] = self.subtype._to_java(value[i])
        return j_arr

    def _from_java(self, value):
        py_array = []
        for j_obj in value:
            py_array.append(self.subtype.from_java(j_obj))
        return py_array


class BigDecimalType(AttributeType):
    """
    Conversion class for BigDecimal.
    """

    def __init__(self):
        super().__init__("java.math.BigDecimal", Decimal)

    def _to_java(self, value):
        return java_pkg.java.math.BigDecimal(str(value))


class BigDecimalArrayType(ArrayAttributeType):
    """
    Conversion class for BigDecimal[].
    """

    def __init__(self):
        super().__init__(BigDecimalType())


class BigIntegerType(AttributeType):
    """
    Conversion class for BigInteger.
    """

    def __init__(self):
        super().__init__("java.math.BigInteger", int)

    def _to_java(self, value):
        return java_pkg.java.math.BigInteger(str(value))

    def _from_java(self, value):
        return int(value.toString())


class BigIntegerArrayType(ArrayAttributeType):
    """
    Conversion class for BigInteger[].
    """

    def __init__(self):
        super().__init__(BigIntegerType())


class BooleanType(AttributeType):
    """
    Conversion class for Boolean.
    """

    def __init__(self):
        super().__init__("java.lang.Boolean", bool)


class BooleanArrayType(ArrayAttributeType):
    """
    Conversion class for Boolean[].
    """

    def __init__(self):
        super().__init__(BooleanType())


class FloatType(AttributeType):
    """
    Conversion class for Float.
    """

    def __init__(self):
        super().__init__("java.lang.Float", (int, float))

    def _to_java(self, value):
        return value * 1.0


class FloatArrayType(ArrayAttributeType):
    """
    Conversion class for Float[].
    """

    def __init__(self):
        super().__init__(FloatType())


class DoubleType(AttributeType):
    """
    Conversion class for Double.
    """

    def __init__(self):
        super().__init__("java.lang.Double", (int, float))

    def _to_java(self, value):
        return value * 1.0


class DoubleArrayType(ArrayAttributeType):
    """
    Conversion class for Double[].
    """

    def __init__(self):
        super().__init__(DoubleType())


class ByteType(AttributeType):
    """
    Conversion class for Byte.
    """

    def __init__(self):
        super().__init__("java.lang.Byte", int)


class ByteArrayType(ArrayAttributeType):
    """
    Conversion class for Byte[].
    """

    def __init__(self):
        super().__init__(ByteType(), (list, bytes))

    def _from_java(self, value):
        if None in value:
            return super()._from_java(value)
        return bytes(super()._from_java(value))


class ShortType(AttributeType):
    """
    Conversion class for Short.
    """

    def __init__(self):
        super().__init__("java.lang.Short", int)


class ShortArrayType(ArrayAttributeType):
    """
    Conversion class for Short[].
    """

    def __init__(self):
        super().__init__(ShortType())


class IntegerType(AttributeType):
    """
    Conversion class for Integer.
    """

    def __init__(self):
        super().__init__("java.lang.Integer", int)


class IntegerArrayType(ArrayAttributeType):
    """
    Conversion class for Integer[].
    """

    def __init__(self):
        super().__init__(IntegerType())


class LongType(AttributeType):
    """
    Conversion class for Long.
    """

    def __init__(self):
        super().__init__("java.lang.Long", int)


class LongArrayType(ArrayAttributeType):
    """
    Conversion class for Long[].
    """

    def __init__(self):
        super().__init__(LongType())


class PrimitiveArrayType(AttributeType):
    """
    Base class for arrays made up of Java primitives.
    """

    def __init__(self, binding, j_class, py_element_type, py_type=list):
        super().__init__(binding, py_type)
        self._j_class = j_class
        self._py_element_type = py_element_type

    def _to_java(self, value):
        j_arr = self._build_array(value)
        for i in range(len(value)):
            if value[i] is None:
                raise ValueError("Value at index %d should not be None for primitive array." % i)
            if isinstance(value[i], self._py_element_type):
                j_arr[i] = self._value_to_java(value[i])
            else:
                self._value_error(value[i], _type_to_string(self._py_element_type))
        return j_arr

    def _from_java(self, value):
        py_array = []
        for j_obj in value:
            py_array.append(self._value_from_java(j_obj))
        return py_array

    def _build_array(self, value):
        return java_gateway.new_array(self._j_class, len(value))

    def _value_to_java(self, value):
        return value

    def _value_from_java(self, value):
        return value


class PrimitiveBooleanArrayType(PrimitiveArrayType):
    """
    Conversion class for boolean[].
    """

    def __init__(self):
        super().__init__("[Z", java_gateway.jvm.boolean, bool)


class PrimitiveFloatArrayType(PrimitiveArrayType):
    """
    Conversion class for float[].
    """

    def __init__(self):
        super().__init__("[F", java_gateway.jvm.float, (int, float))

    def _value_to_java(self, value):
        return value * 1.0


class PrimitiveDoubleArrayType(PrimitiveArrayType):
    """
    Conversion class for double[].
    """

    def __init__(self):
        super().__init__("[D", java_gateway.jvm.double, (int, float))

    def _value_to_java(self, value):
        return value * 1.0


class PrimitiveByteArrayType(PrimitiveArrayType):
    """
    Conversion class for byte[].
    """

    def __init__(self):
        super().__init__("[B", java_gateway.jvm.byte, int, (list, bytes))

    def _to_java(self, value):
        if isinstance(value, bytes):
            return value
        return bytes(super()._to_java(value))

    def _from_java(self, value):
        if isinstance(value, bytes):
            return value
        return bytes(super()._from_java(value))

    def _value_to_java(self, value):
        return value % 256

    def _build_array(self, value):
        return bytearray(super()._build_array(value))


class PrimitiveShortArrayType(PrimitiveArrayType):
    """
    Conversion class for short[].
    """

    def __init__(self):
        super().__init__("[S", java_gateway.jvm.short, int)


class PrimitiveIntArrayType(PrimitiveArrayType):
    """
    Conversion class for int[].
    """

    def __init__(self):
        super().__init__("[I", java_gateway.jvm.int, int)


class PrimitiveLongArrayType(PrimitiveArrayType):
    """
    Conversion class for long[].
    """

    def __init__(self):
        super().__init__("[J", java_gateway.jvm.long, int)


class StringType(AttributeType):
    """
    Conversion class for String.
    """

    def __init__(self):
        super().__init__("java.lang.String", str)


class StringArrayType(ArrayAttributeType):
    """
    Conversion class for String[].
    """

    def __init__(self):
        super().__init__(StringType())


class DateType(AttributeType):
    """
    Conversion class for Date.
    """

    def __init__(self):
        super().__init__("java.util.Date", date)

    def _to_java(self, value):
        return java_pkg.java.util.Date(int(value.replace(tzinfo=timezone.utc).timestamp() * 1000))

    def _from_java(self, value):
        return datetime.utcfromtimestamp(value.getTime() / 1000)


class DateArrayType(ArrayAttributeType):
    """
    Conversion class for Date[].
    """

    def __init__(self):
        super().__init__(DateType())


class CalendarType(AttributeType):
    """
    Conversion class for Calendar.
    """

    def __init__(self):
        super().__init__("java.util.Calendar", date)
        self._date_type = DateType()

    def _to_java(self, value):
        j_timezone = java_pkg.java.util.TimeZone.getTimeZone("GMT")
        j_calendar = java_pkg.java.util.Calendar.getInstance(j_timezone)
        j_calendar.setTime(self._date_type._to_java(value))
        return j_calendar

    def _from_java(self, value):
        return self._date_type._from_java(value.getTime())


class CalendarArrayType(ArrayAttributeType):
    """
    Conversion class for Calendar[].
    """

    def __init__(self):
        super().__init__(CalendarType())


class BaseGeometryType(AttributeType):
    """
    Base type for conversion between shapely geometries and JTS geometries.
    """

    def __init__(self, binding, py_type):
        super().__init__(binding, py_type)

    def _to_java(self, value):
        wkb = dumps(value)
        return _wkb_reader.read(wkb)

    def _from_java(self, value):
        wkb = _wkb_writer.write(value)
        return loads(wkb)


class PointType(BaseGeometryType):
    """
    Conversion class for Point.
    """

    def __init__(self):
        super().__init__("org.locationtech.jts.geom.Point", Point)


class PointArrayType(ArrayAttributeType):
    """
    Conversion class for Point[].
    """

    def __init__(self):
        super().__init__(PointType())


class MultiPointType(BaseGeometryType):
    """
    Conversion class for MultiPoint.
    """

    def __init__(self):
        super().__init__("org.locationtech.jts.geom.MultiPoint", MultiPoint)


class MultiPointArrayType(ArrayAttributeType):
    """
    Conversion class for MultiPoint[].
    """

    def __init__(self):
        super().__init__(MultiPointType())


class LineStringType(BaseGeometryType):
    """
    Conversion class for LineString.
    """

    def __init__(self):
        super().__init__("org.locationtech.jts.geom.LineString", LineString)


class LineStringArrayType(ArrayAttributeType):
    """
    Conversion class for LineString[].
    """

    def __init__(self):
        super().__init__(LineStringType())


class MultiLineStringType(BaseGeometryType):
    """
    Conversion class for MultiLineString.
    """

    def __init__(self):
        super().__init__("org.locationtech.jts.geom.MultiLineString", MultiLineString)


class MultiLineStringArrayType(ArrayAttributeType):
    """
    Conversion class for MultiLineString[].
    """

    def __init__(self):
        super().__init__(MultiLineStringType())


class PolygonType(BaseGeometryType):
    """
    Conversion class for Polygon.
    """

    def __init__(self):
        super().__init__("org.locationtech.jts.geom.Polygon", Polygon)


class PolygonArrayType(ArrayAttributeType):
    """
    Conversion class for Polygon[].
    """

    def __init__(self):
        super().__init__(PolygonType())


class MultiPolygonType(BaseGeometryType):
    """
    Conversion class for MultiPolygon.
    """

    def __init__(self):
        super().__init__("org.locationtech.jts.geom.MultiPolygon", MultiPolygon)


class MultiPolygonArrayType(ArrayAttributeType):
    """
    Conversion class for MultiPolygon[].
    """

    def __init__(self):
        super().__init__(MultiPolygonType())


class GeometryCollectionType(BaseGeometryType):
    """
    Conversion class for GeometryCollection.
    """

    def __init__(self):
        super().__init__("org.locationtech.jts.geom.GeometryCollection", GeometryCollection)


class GeometryCollectionArrayType(ArrayAttributeType):
    """
    Conversion class for GeometryCollection[].
    """

    def __init__(self):
        super().__init__(GeometryCollectionType())


class GeometryType(BaseGeometryType):
    """
    Conversion class for Geometry.
    """

    def __init__(self):
        super().__init__("org.locationtech.jts.geom.Geometry", BaseGeometry)


class GeometryArrayType(ArrayAttributeType):
    """
    Conversion class for Geometry[].
    """

    def __init__(self):
        super().__init__(GeometryType())



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.config import geowave_pkg
from ..statistic_binning_strategy import StatisticBinningStrategy


class PartitionBinningStrategy(StatisticBinningStrategy):
    """
    Statistic binning strategy that bins statistic values by the partitions that the entry resides on.
    """

    def __init__(self, java_ref=None):
        if java_ref is None:
            java_ref = geowave_pkg.core.store.statistics.binning.PartitionBinningStrategy()
        super().__init__(java_ref)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from .geowave_object import GeoWaveObject
from .write_results import WriteResults


class Writer(GeoWaveObject):
    """
    Writes data to a GeoWave data store.
    """

    def __init__(self, java_ref):
        super().__init__(java_ref)
        self.is_open = True

    def write(self, data):
        """
        Write data into the associated data store.

        Args:
            data (any) : The data to be written.
        Raises:
            RuntimeError: If the writer is closed.
        Returns:
            A `pygw.base.write_results.WriteResults` with the results of the write operation.
        """
        if not self.is_open:
            raise RuntimeError("Writer is already closed!")

        if isinstance(data, GeoWaveObject):
            data = data._java_ref

        return WriteResults(self._java_ref.write(data))

    def close(self):
        """
        Close the writer.
        """
        if self.is_open:
            self._java_ref.close()
            self.is_open = False



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
"""
This module contains the classes needed for querying statistics.
"""

from .statistic_query_builder import StatisticQueryBuilder
from .statistic_query_builder import DataTypeStatisticQueryBuilder
from .statistic_query_builder import FieldStatisticQueryBuilder
from .statistic_query_builder import IndexStatisticQueryBuilder
from .statistic_query import StatisticQuery



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base.type_conversions import StringArrayType
from pygw.config import geowave_pkg
from .field_value_binning_strategy import FieldValueBinningStrategy


class NumericRangeFieldValueBinningStrategy(FieldValueBinningStrategy):
    """
    Statistic binning strategy that bins statistic values by the numeric representation of the value of a given field.
    By default it will truncate decimal places and will bin by the integer. However, an "offset" and "interval" can be
    provided to bin numbers at any regular step-sized increment from an origin value. A statistic using this binning
    strategy can be constrained using numeric ranges (A Range can be used as a constraint).
    """

    def __init__(self, fields=None, interval=1, offset=0, java_ref=None):
        if java_ref is None:
            java_ref = geowave_pkg.core.store.statistics.binning.NumericRangeFieldValueBinningStrategy(
                float(interval), float(offset), StringArrayType().to_java(fields))
        super().__init__(fields, java_ref)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from datetime import datetime
from functools import reduce

from pygw.index import SpatialIndexBuilder

from .conftest import POINT_TYPE_ADAPTER, POINT_TYPE_NAME, POINT_NUMBER_FIELD, POINT_GEOMETRY_FIELD, POINT_TIME_FIELD, \
    POINT_SHAPE_FIELD, POINT_COLOR_FIELD, results_as_list
from .conftest import write_test_data, write_test_data_offset
from ..base import Interval, Envelope

from shapely.geometry import Polygon
from ..base.range import Range
from ..query.statistics.statistic_query_builder import DataTypeStatisticQueryBuilder
from ..query.statistics.statistic_query_builder import FieldStatisticQueryBuilder
from ..query.statistics.statistic_query_builder import IndexStatisticQueryBuilder
from ..query.statistics.statistic_query_builder import StatisticQueryBuilder
from ..query.statistics.statistic_query import StatisticQuery
from ..statistics import DataTypeStatisticType, FieldStatisticType, IndexStatisticType, BinConstraints, StatisticValue
from ..statistics.binning_strategy import CompositeBinningStrategy, DataTypeBinningStrategy, \
    FieldValueBinningStrategy, NumericRangeFieldValueBinningStrategy, TimeRangeFieldValueBinningStrategy, \
    PartitionBinningStrategy, SpatialFieldValueBinningStrategy
from ..statistics.data_type import CountStatistic
from ..statistics.field import BloomFilterStatistic, BoundingBoxStatistic, CountMinSketchStatistic, \
    FixedBinNumericHistogramStatistic, HyperLogLogStatistic, NumericHistogramStatistic, NumericMeanStatistic, \
    NumericRangeStatistic, NumericStatsStatistic, TimeRangeStatistic
from ..statistics.field.bloom_filter_statistic import BloomFilter
from ..statistics.field.count_min_sketch_statistic import CountMinSketch
from ..statistics.field.fixed_bin_numeric_histogram_statistic import FixedBinNumericHistogram
from ..statistics.field.hyper_log_log_statistic import HyperLogLogPlus
from ..statistics.field.numeric_histogram_statistic import NumericHistogram
from ..statistics.field.numeric_stats_statistic import Stats
from ..statistics.index import DifferingVisibilityCountStatistic, DuplicateEntryCountStatistic, \
    FieldVisibilityCountStatistic, MaxDuplicatesStatistic, PartitionsStatistic, RowRangeHistogramStatistic, \
    IndexMetaDataSetStatistic

INTERNAL_TAG = 'internal'
TEST_TAG = 'test'
TEST_INDEX = 'spatial_index'


def test_statistic_query(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    binning_strategy = FieldValueBinningStrategy([POINT_COLOR_FIELD, POINT_SHAPE_FIELD])
    count_stat = CountStatistic()
    count_stat.set_tag(TEST_TAG)
    count_stat.set_type_name(POINT_TYPE_NAME)
    count_stat.set_binning_strategy(binning_strategy)
    test_ds.add_statistic(count_stat)
    write_test_data(test_ds, index)

    # when
    stat_query_builder = StatisticQueryBuilder.new_builder(CountStatistic.STATS_TYPE)
    stat_query_builder.type_name(POINT_TYPE_NAME)
    stat_query_builder.tag(TEST_TAG)
    stat_query_builder.bin_constraints(BinConstraints.of_prefix(b'BLUE'))
    result = results_as_list(test_ds.query_statistics(stat_query_builder.build()))

    # then
    assert len(result) == 4
    merged = None
    for stat_value in result:
        if merged is None:
            merged = stat_value
        else:
            merged.merge(stat_value)
        assert isinstance(stat_value, StatisticValue)
        assert stat_value.get_value() == 30
        assert isinstance(stat_value.get_statistic(), CountStatistic)
    assert merged.get_value() == 120

    # when
    aggregated_result = test_ds.aggregate_statistics(stat_query_builder.build())

    # then
    assert isinstance(aggregated_result, StatisticValue)
    assert aggregated_result.get_value() == 120


def test_query_transform(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)
    numeric_range_stat = NumericRangeStatistic()
    numeric_range_stat.set_tag(TEST_TAG)
    numeric_range_stat.set_type_name(POINT_TYPE_NAME)
    numeric_range_stat.set_field_name(POINT_NUMBER_FIELD)
    test_ds.add_statistic(numeric_range_stat)
    write_test_data(test_ds, index)

    # when
    stat_query_builder = StatisticQueryBuilder.new_builder(NumericRangeStatistic.STATS_TYPE)
    stat_query_builder.type_name(POINT_TYPE_NAME)
    stat_query_builder.field_name(POINT_NUMBER_FIELD)
    stat_query_builder.tag(TEST_TAG)
    result = results_as_list(test_ds.query_statistics(stat_query_builder.build()))

    # then
    assert len(result) == 1
    assert isinstance(result[0], StatisticValue)
    value = result[0].get_value()
    assert isinstance(value, Range)
    assert value.get_minimum() == -180
    assert value.get_maximum() == 179

    # when
    aggregated_result = test_ds.aggregate_statistics(stat_query_builder.build())

    # then
    assert isinstance(aggregated_result, StatisticValue)
    value = aggregated_result.get_value()
    assert isinstance(value, Range)
    assert value.get_minimum() == -180
    assert value.get_maximum() == 179


def test_statistic_query_builders():
    # exercise all methods to make sure their java connections are valid
    index_query_builder = StatisticQueryBuilder.new_builder(DifferingVisibilityCountStatistic.STATS_TYPE)
    assert isinstance(index_query_builder, IndexStatisticQueryBuilder)
    index_query_builder.index_name('idx')
    index_query_builder.tag('test')
    index_query_builder.add_authorization('auth1')
    index_query_builder.authorizations(['auth1', 'auth2'])
    index_query_builder.internal()
    query = index_query_builder.build()
    assert isinstance(query, StatisticQuery)

    data_type_query_builder = StatisticQueryBuilder.new_builder(CountStatistic.STATS_TYPE)
    assert isinstance(data_type_query_builder, DataTypeStatisticQueryBuilder)
    data_type_query_builder.type_name(POINT_TYPE_NAME)
    query = data_type_query_builder.build()
    assert isinstance(query, StatisticQuery)

    field_query_builder = StatisticQueryBuilder.new_builder(BloomFilterStatistic.STATS_TYPE)
    assert isinstance(field_query_builder, FieldStatisticQueryBuilder)
    field_query_builder.type_name(POINT_TYPE_NAME)
    field_query_builder.field_name(POINT_NUMBER_FIELD)
    query = field_query_builder.build()
    assert isinstance(query, StatisticQuery)

    statistic_query_builder = StatisticQueryBuilder.count()
    assert isinstance(statistic_query_builder, DataTypeStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.bloom_filter()
    assert isinstance(statistic_query_builder, FieldStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.bbox()
    assert isinstance(statistic_query_builder, FieldStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.count_min_sketch()
    assert isinstance(statistic_query_builder, FieldStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.fixed_bin_numeric_histogram()
    assert isinstance(statistic_query_builder, FieldStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.hyper_log_log()
    assert isinstance(statistic_query_builder, FieldStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.numeric_histogram()
    assert isinstance(statistic_query_builder, FieldStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.numeric_mean()
    assert isinstance(statistic_query_builder, FieldStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.numeric_range()
    assert isinstance(statistic_query_builder, FieldStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.differing_visibility_count()
    assert isinstance(statistic_query_builder, IndexStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.duplicate_entry_count()
    assert isinstance(statistic_query_builder, IndexStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.field_visibility_count()
    assert isinstance(statistic_query_builder, IndexStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.index_meta_data_set()
    assert isinstance(statistic_query_builder, IndexStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.max_duplicates()
    assert isinstance(statistic_query_builder, IndexStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.partitions()
    assert isinstance(statistic_query_builder, IndexStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.row_range_histogram()
    assert isinstance(statistic_query_builder, IndexStatisticQueryBuilder)

    statistic_query_builder = StatisticQueryBuilder.numeric_stats()
    assert isinstance(statistic_query_builder, FieldStatisticQueryBuilder)


def test_count_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    count_stat = CountStatistic()
    count_stat.set_tag(TEST_TAG)
    count_stat.set_type_name(POINT_TYPE_NAME)
    test_ds.add_statistic(count_stat)
    write_test_data(test_ds, index)

    # then
    count_stat = test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, TEST_TAG)
    assert isinstance(count_stat, CountStatistic)
    assert count_stat.get_tag() == TEST_TAG
    assert count_stat.get_type_name() == POINT_TYPE_NAME
    assert count_stat.get_description() is not None
    assert count_stat.get_binning_strategy() is None
    assert isinstance(count_stat.get_statistic_type(), DataTypeStatisticType)
    assert count_stat.get_statistic_type().get_string() == 'COUNT'
    assert test_ds.get_statistic_value(count_stat) == 360

    # test alternate constructors
    count_stat = CountStatistic(POINT_TYPE_NAME)
    count_stat.set_internal()
    assert count_stat.get_type_name() == POINT_TYPE_NAME
    assert count_stat.get_tag() == INTERNAL_TAG


def test_bloom_filter_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    bloom_filter_stat = BloomFilterStatistic()
    bloom_filter_stat.set_tag(TEST_TAG)
    bloom_filter_stat.set_type_name(POINT_TYPE_NAME)
    bloom_filter_stat.set_field_name(POINT_NUMBER_FIELD)
    test_ds.add_statistic(bloom_filter_stat)
    write_test_data(test_ds, index)

    # then
    bloom_filter_stat = test_ds.get_field_statistic(
        BloomFilterStatistic.STATS_TYPE,
        POINT_TYPE_NAME,
        POINT_NUMBER_FIELD,
        TEST_TAG)
    assert isinstance(bloom_filter_stat, BloomFilterStatistic)
    assert bloom_filter_stat.get_tag() == TEST_TAG
    assert bloom_filter_stat.get_type_name() == POINT_TYPE_NAME
    assert bloom_filter_stat.get_field_name() == POINT_NUMBER_FIELD
    assert bloom_filter_stat.get_description() is not None
    assert bloom_filter_stat.get_binning_strategy() is None
    assert isinstance(bloom_filter_stat.get_statistic_type(), FieldStatisticType)
    assert bloom_filter_stat.get_statistic_type().get_string() == 'BLOOM_FILTER'
    bloom_filter = test_ds.get_statistic_value(bloom_filter_stat)
    assert isinstance(bloom_filter, BloomFilter)
    assert bloom_filter.might_contain('-180.0')
    assert bloom_filter.might_contain('179.0')
    assert not bloom_filter.might_contain('garbage')
    assert bloom_filter.approximate_element_count() > 300
    assert bloom_filter.expected_fpp() is not None

    # test alternate constructors
    bloom_filter_stat = BloomFilterStatistic(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    bloom_filter_stat.set_internal()
    assert bloom_filter_stat.get_type_name() == POINT_TYPE_NAME
    assert bloom_filter_stat.get_field_name() == POINT_NUMBER_FIELD
    assert bloom_filter_stat.get_tag() == INTERNAL_TAG


def test_bounding_box_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    bounding_box_stat = BoundingBoxStatistic()
    bounding_box_stat.set_tag(TEST_TAG)
    bounding_box_stat.set_type_name(POINT_TYPE_NAME)
    bounding_box_stat.set_field_name(POINT_GEOMETRY_FIELD)
    test_ds.add_statistic(bounding_box_stat)
    write_test_data_offset(test_ds, index)

    # then
    bounding_box_stat = test_ds.get_field_statistic(
        BoundingBoxStatistic.STATS_TYPE,
        POINT_TYPE_NAME,
        POINT_GEOMETRY_FIELD,
        TEST_TAG)
    assert isinstance(bounding_box_stat, BoundingBoxStatistic)
    assert bounding_box_stat.get_tag() == TEST_TAG
    assert bounding_box_stat.get_type_name() == POINT_TYPE_NAME
    assert bounding_box_stat.get_field_name() == POINT_GEOMETRY_FIELD
    assert bounding_box_stat.get_description() is not None
    assert bounding_box_stat.get_binning_strategy() is None
    assert isinstance(bounding_box_stat.get_statistic_type(), FieldStatisticType)
    assert bounding_box_stat.get_statistic_type().get_string() == 'BOUNDING_BOX'
    bounding_box = test_ds.get_statistic_value(bounding_box_stat)
    assert isinstance(bounding_box, Envelope)
    assert bounding_box.get_min_x() == -179.5
    assert bounding_box.get_min_y() == -89.5
    assert bounding_box.get_max_x() == 179.5
    assert bounding_box.get_max_y() == 89.5

    # test alternate constructors
    bounding_box_stat = BoundingBoxStatistic(POINT_TYPE_NAME, POINT_GEOMETRY_FIELD)
    bounding_box_stat.set_internal()
    assert bounding_box_stat.get_type_name() == POINT_TYPE_NAME
    assert bounding_box_stat.get_field_name() == POINT_GEOMETRY_FIELD
    assert bounding_box_stat.get_tag() == INTERNAL_TAG


def test_count_min_sketch_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    count_min_sketch_stat = CountMinSketchStatistic()
    count_min_sketch_stat.set_tag(TEST_TAG)
    count_min_sketch_stat.set_type_name(POINT_TYPE_NAME)
    count_min_sketch_stat.set_field_name(POINT_NUMBER_FIELD)
    count_min_sketch_stat.set_error_factor(0.002)
    count_min_sketch_stat.set_probability_of_correctness(0.8)
    test_ds.add_statistic(count_min_sketch_stat)
    write_test_data(test_ds, index)

    # then
    count_min_sketch_stat = test_ds.get_field_statistic(
        CountMinSketchStatistic.STATS_TYPE,
        POINT_TYPE_NAME,
        POINT_NUMBER_FIELD,
        TEST_TAG)
    assert isinstance(count_min_sketch_stat, CountMinSketchStatistic)
    assert count_min_sketch_stat.get_tag() == TEST_TAG
    assert count_min_sketch_stat.get_type_name() == POINT_TYPE_NAME
    assert count_min_sketch_stat.get_field_name() == POINT_NUMBER_FIELD
    assert count_min_sketch_stat.get_error_factor() == 0.002
    assert count_min_sketch_stat.get_probability_of_correctness() == 0.8
    assert count_min_sketch_stat.get_description() is not None
    assert count_min_sketch_stat.get_binning_strategy() is None
    assert isinstance(count_min_sketch_stat.get_statistic_type(), FieldStatisticType)
    assert count_min_sketch_stat.get_statistic_type().get_string() == 'COUNT_MIN_SKETCH'
    count_min_sketch = test_ds.get_statistic_value(count_min_sketch_stat)
    assert isinstance(count_min_sketch, CountMinSketch)
    assert count_min_sketch.get_relative_error() == 0.002
    assert count_min_sketch.get_confidence() > 0.5
    assert count_min_sketch.estimate_count('90.0') == 1

    # test alternate constructors
    count_min_sketch = CountMinSketchStatistic(POINT_TYPE_NAME, POINT_GEOMETRY_FIELD, 0.002, 0.8)
    count_min_sketch.set_internal()
    assert count_min_sketch.get_type_name() == POINT_TYPE_NAME
    assert count_min_sketch.get_field_name() == POINT_GEOMETRY_FIELD
    assert count_min_sketch.get_error_factor() == 0.002
    assert count_min_sketch.get_probability_of_correctness() == 0.8
    assert count_min_sketch.get_tag() == INTERNAL_TAG


def test_fixed_bin_numeric_histogram(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    fixed_bin_stat = FixedBinNumericHistogramStatistic()
    fixed_bin_stat.set_tag(TEST_TAG)
    fixed_bin_stat.set_type_name(POINT_TYPE_NAME)
    fixed_bin_stat.set_field_name(POINT_NUMBER_FIELD)
    fixed_bin_stat.set_num_bins(128)
    fixed_bin_stat.set_min_value(-180.0)
    fixed_bin_stat.set_max_value(180.0)
    test_ds.add_statistic(fixed_bin_stat)
    write_test_data(test_ds, index)

    # then
    fixed_bin_stat = test_ds.get_field_statistic(
        FixedBinNumericHistogramStatistic.STATS_TYPE,
        POINT_TYPE_NAME,
        POINT_NUMBER_FIELD,
        TEST_TAG)
    assert isinstance(fixed_bin_stat, FixedBinNumericHistogramStatistic)
    assert fixed_bin_stat.get_tag() == TEST_TAG
    assert fixed_bin_stat.get_type_name() == POINT_TYPE_NAME
    assert fixed_bin_stat.get_field_name() == POINT_NUMBER_FIELD
    assert fixed_bin_stat.get_num_bins() == 128
    assert fixed_bin_stat.get_min_value() == -180
    assert fixed_bin_stat.get_max_value() == 180
    assert fixed_bin_stat.get_description() is not None
    assert fixed_bin_stat.get_binning_strategy() is None
    assert isinstance(fixed_bin_stat.get_statistic_type(), FieldStatisticType)
    assert fixed_bin_stat.get_statistic_type().get_string() == 'FIXED_BIN_NUMERIC_HISTOGRAM'
    histogram = test_ds.get_statistic_value(fixed_bin_stat)
    assert isinstance(histogram, FixedBinNumericHistogram)
    assert histogram.get_num_bins() == 128
    assert histogram.get_min_value() == -180
    assert histogram.get_max_value() == 180
    assert histogram.cdf(0) == 0.5
    assert len(histogram.bin_quantiles(0)) == 0
    assert histogram.quantile(0.0) == -180
    assert 180 < histogram.quantile(1.0) < 185
    assert histogram.sum(0) == 180
    assert histogram.percent_population_over_range(-90, 0) == 0.25
    assert histogram.total_sample_size() == 360
    assert histogram.get_total_count() == 360
    assert len(histogram.count(0)) == 0

    # test alternate constructors
    fixed_bin_stat = FixedBinNumericHistogramStatistic(POINT_TYPE_NAME, POINT_NUMBER_FIELD, 128, -180, 180)
    fixed_bin_stat.set_internal()
    assert fixed_bin_stat.get_type_name() == POINT_TYPE_NAME
    assert fixed_bin_stat.get_field_name() == POINT_NUMBER_FIELD
    assert fixed_bin_stat.get_num_bins() == 128
    assert fixed_bin_stat.get_min_value() == -180
    assert fixed_bin_stat.get_max_value() == 180
    assert fixed_bin_stat.get_tag() == INTERNAL_TAG


def test_hyper_log_log_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    hyper_log_log_stat = HyperLogLogStatistic()
    hyper_log_log_stat.set_tag(TEST_TAG)
    hyper_log_log_stat.set_type_name(POINT_TYPE_NAME)
    hyper_log_log_stat.set_field_name(POINT_NUMBER_FIELD)
    hyper_log_log_stat.set_precision(24)
    test_ds.add_statistic(hyper_log_log_stat)
    write_test_data(test_ds, index)

    # then
    hyper_log_log_stat = test_ds.get_field_statistic(
        HyperLogLogStatistic.STATS_TYPE,
        POINT_TYPE_NAME,
        POINT_NUMBER_FIELD,
        TEST_TAG)
    assert isinstance(hyper_log_log_stat, HyperLogLogStatistic)
    assert hyper_log_log_stat.get_tag() == TEST_TAG
    assert hyper_log_log_stat.get_type_name() == POINT_TYPE_NAME
    assert hyper_log_log_stat.get_field_name() == POINT_NUMBER_FIELD
    assert hyper_log_log_stat.get_precision() == 24
    assert hyper_log_log_stat.get_description() is not None
    assert hyper_log_log_stat.get_binning_strategy() is None
    assert isinstance(hyper_log_log_stat.get_statistic_type(), FieldStatisticType)
    assert hyper_log_log_stat.get_statistic_type().get_string() == 'HYPER_LOG_LOG'
    hyper_log_log = test_ds.get_statistic_value(hyper_log_log_stat)
    assert isinstance(hyper_log_log, HyperLogLogPlus)
    assert 350 < hyper_log_log.cardinality() < 370

    # test alternate constructors
    fixed_bin_stat = HyperLogLogStatistic(POINT_TYPE_NAME, POINT_NUMBER_FIELD, 31)
    fixed_bin_stat.set_internal()
    assert fixed_bin_stat.get_type_name() == POINT_TYPE_NAME
    assert fixed_bin_stat.get_field_name() == POINT_NUMBER_FIELD
    assert fixed_bin_stat.get_precision() == 31
    assert fixed_bin_stat.get_tag() == INTERNAL_TAG


def test_numeric_histogram_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    numeric_histogram_stat = NumericHistogramStatistic()
    numeric_histogram_stat.set_tag(TEST_TAG)
    numeric_histogram_stat.set_type_name(POINT_TYPE_NAME)
    numeric_histogram_stat.set_field_name(POINT_NUMBER_FIELD)
    numeric_histogram_stat.set_compression(80)
    test_ds.add_statistic(numeric_histogram_stat)
    write_test_data(test_ds, index)

    # then
    numeric_histogram_stat = test_ds.get_field_statistic(
        NumericHistogramStatistic.STATS_TYPE,
        POINT_TYPE_NAME,
        POINT_NUMBER_FIELD,
        TEST_TAG)
    assert isinstance(numeric_histogram_stat, NumericHistogramStatistic)
    assert numeric_histogram_stat.get_tag() == TEST_TAG
    assert numeric_histogram_stat.get_type_name() == POINT_TYPE_NAME
    assert numeric_histogram_stat.get_field_name() == POINT_NUMBER_FIELD
    assert numeric_histogram_stat.get_compression() == 80
    assert numeric_histogram_stat.get_description() is not None
    assert numeric_histogram_stat.get_binning_strategy() is None
    assert isinstance(numeric_histogram_stat.get_statistic_type(), FieldStatisticType)
    assert numeric_histogram_stat.get_statistic_type().get_string() == 'NUMERIC_HISTOGRAM'
    histogram = test_ds.get_statistic_value(numeric_histogram_stat)
    assert isinstance(histogram, NumericHistogram)
    assert -182 < histogram.get_min_value() < -178
    assert 178 < histogram.get_max_value() < 182
    assert 0.48 < histogram.cdf(0) < 0.52
    assert -182 < histogram.quantile(0.0) < -178
    assert 178 < histogram.quantile(1.0) < 182
    assert 178 < histogram.sum(0) < 182
    assert histogram.get_total_count() == 360

    # test alternate constructors
    numeric_histogram_stat = NumericHistogramStatistic(POINT_TYPE_NAME, POINT_NUMBER_FIELD, 80)
    numeric_histogram_stat.set_internal()
    assert numeric_histogram_stat.get_type_name() == POINT_TYPE_NAME
    assert numeric_histogram_stat.get_field_name() == POINT_NUMBER_FIELD
    assert numeric_histogram_stat.get_compression() == 80
    assert numeric_histogram_stat.get_tag() == INTERNAL_TAG


def test_numeric_mean_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    numeric_mean_stat = NumericMeanStatistic()
    numeric_mean_stat.set_tag(TEST_TAG)
    numeric_mean_stat.set_type_name(POINT_TYPE_NAME)
    numeric_mean_stat.set_field_name(POINT_NUMBER_FIELD)
    test_ds.add_statistic(numeric_mean_stat)
    write_test_data(test_ds, index)

    # then
    numeric_mean_stat = test_ds.get_field_statistic(
        NumericMeanStatistic.STATS_TYPE,
        POINT_TYPE_NAME,
        POINT_NUMBER_FIELD,
        TEST_TAG)
    assert isinstance(numeric_mean_stat, NumericMeanStatistic)
    assert numeric_mean_stat.get_tag() == TEST_TAG
    assert numeric_mean_stat.get_type_name() == POINT_TYPE_NAME
    assert numeric_mean_stat.get_field_name() == POINT_NUMBER_FIELD
    assert numeric_mean_stat.get_description() is not None
    assert numeric_mean_stat.get_binning_strategy() is None
    assert isinstance(numeric_mean_stat.get_statistic_type(), FieldStatisticType)
    assert numeric_mean_stat.get_statistic_type().get_string() == 'NUMERIC_MEAN'
    assert test_ds.get_statistic_value(numeric_mean_stat) == -0.5  # Mean of values -180 to 179

    # test alternate constructors
    numeric_mean_stat = NumericMeanStatistic(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    numeric_mean_stat.set_internal()
    assert numeric_mean_stat.get_type_name() == POINT_TYPE_NAME
    assert numeric_mean_stat.get_field_name() == POINT_NUMBER_FIELD
    assert numeric_mean_stat.get_tag() == INTERNAL_TAG


def test_numeric_range_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    numeric_range_stat = NumericRangeStatistic()
    numeric_range_stat.set_tag(TEST_TAG)
    numeric_range_stat.set_type_name(POINT_TYPE_NAME)
    numeric_range_stat.set_field_name(POINT_NUMBER_FIELD)
    test_ds.add_statistic(numeric_range_stat)
    write_test_data(test_ds, index)

    # then
    numeric_range_stat = test_ds.get_field_statistic(
        NumericRangeStatistic.STATS_TYPE,
        POINT_TYPE_NAME,
        POINT_NUMBER_FIELD,
        TEST_TAG)
    assert isinstance(numeric_range_stat, NumericRangeStatistic)
    assert numeric_range_stat.get_tag() == TEST_TAG
    assert numeric_range_stat.get_type_name() == POINT_TYPE_NAME
    assert numeric_range_stat.get_field_name() == POINT_NUMBER_FIELD
    assert numeric_range_stat.get_description() is not None
    assert numeric_range_stat.get_binning_strategy() is None
    assert isinstance(numeric_range_stat.get_statistic_type(), FieldStatisticType)
    assert numeric_range_stat.get_statistic_type().get_string() == 'NUMERIC_RANGE'
    numeric_range = test_ds.get_statistic_value(numeric_range_stat)
    assert isinstance(numeric_range, Range)
    assert numeric_range.get_minimum() == -180
    assert numeric_range.get_maximum() == 179

    # test alternate constructors
    numeric_range_stat = NumericRangeStatistic(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    numeric_range_stat.set_internal()
    assert numeric_range_stat.get_type_name() == POINT_TYPE_NAME
    assert numeric_range_stat.get_field_name() == POINT_NUMBER_FIELD
    assert numeric_range_stat.get_tag() == INTERNAL_TAG


def test_numeric_stats_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    numeric_stats_stat = NumericStatsStatistic()
    numeric_stats_stat.set_tag(TEST_TAG)
    numeric_stats_stat.set_type_name(POINT_TYPE_NAME)
    numeric_stats_stat.set_field_name(POINT_NUMBER_FIELD)
    test_ds.add_statistic(numeric_stats_stat)
    write_test_data(test_ds, index)

    # then
    numeric_stats_stat = test_ds.get_field_statistic(
        NumericStatsStatistic.STATS_TYPE,
        POINT_TYPE_NAME,
        POINT_NUMBER_FIELD,
        TEST_TAG)
    assert isinstance(numeric_stats_stat, NumericStatsStatistic)
    assert numeric_stats_stat.get_tag() == TEST_TAG
    assert numeric_stats_stat.get_type_name() == POINT_TYPE_NAME
    assert numeric_stats_stat.get_field_name() == POINT_NUMBER_FIELD
    assert numeric_stats_stat.get_description() is not None
    assert numeric_stats_stat.get_binning_strategy() is None
    assert isinstance(numeric_stats_stat.get_statistic_type(), FieldStatisticType)
    assert numeric_stats_stat.get_statistic_type().get_string() == 'NUMERIC_STATS'
    stats = test_ds.get_statistic_value(numeric_stats_stat)
    assert isinstance(stats, Stats)
    assert stats.count() == 360
    assert stats.mean() == -0.5
    assert stats.sum() == -180
    assert 10799 < stats.population_variance() < 10801
    assert 103 < stats.population_standard_deviation() < 105
    assert 10829 < stats.sample_variance() < 10831
    assert 103 < stats.sample_standard_deviation() < 105
    assert stats.min() == -180
    assert stats.max() == 179

    # test alternate constructors
    numeric_stats_stat = NumericStatsStatistic(POINT_TYPE_NAME, POINT_NUMBER_FIELD)
    numeric_stats_stat.set_internal()
    assert numeric_stats_stat.get_type_name() == POINT_TYPE_NAME
    assert numeric_stats_stat.get_field_name() == POINT_NUMBER_FIELD
    assert numeric_stats_stat.get_tag() == INTERNAL_TAG


def test_time_range_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    time_range_stat = TimeRangeStatistic()
    time_range_stat.set_tag(TEST_TAG)
    time_range_stat.set_type_name(POINT_TYPE_NAME)
    time_range_stat.set_field_name(POINT_TIME_FIELD)
    test_ds.add_statistic(time_range_stat)
    write_test_data(test_ds, index)

    # then
    time_range_stat = test_ds.get_field_statistic(
        TimeRangeStatistic.STATS_TYPE,
        POINT_TYPE_NAME,
        POINT_TIME_FIELD,
        TEST_TAG)
    assert isinstance(time_range_stat, TimeRangeStatistic)
    assert time_range_stat.get_tag() == TEST_TAG
    assert time_range_stat.get_type_name() == POINT_TYPE_NAME
    assert time_range_stat.get_field_name() == POINT_TIME_FIELD
    assert time_range_stat.get_description() is not None
    assert time_range_stat.get_binning_strategy() is None
    assert isinstance(time_range_stat.get_statistic_type(), FieldStatisticType)
    assert time_range_stat.get_statistic_type().get_string() == 'TIME_RANGE'
    time_range = test_ds.get_statistic_value(time_range_stat)
    assert isinstance(time_range, Interval)
    assert time_range.get_start() == datetime.utcfromtimestamp(-180)
    assert time_range.get_end() == datetime.utcfromtimestamp(179)

    # test alternate constructors
    time_range_stat = TimeRangeStatistic(POINT_TYPE_NAME, POINT_TIME_FIELD)
    time_range_stat.set_internal()
    assert time_range_stat.get_type_name() == POINT_TYPE_NAME
    assert time_range_stat.get_field_name() == POINT_TIME_FIELD
    assert time_range_stat.get_tag() == INTERNAL_TAG


def test_differing_visibility_count_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    differing_visibility_stat = DifferingVisibilityCountStatistic()
    differing_visibility_stat.set_tag(TEST_TAG)
    differing_visibility_stat.set_index_name(TEST_INDEX)
    test_ds.add_statistic(differing_visibility_stat)
    write_test_data(test_ds, index)

    # then
    differing_visibility_stat = test_ds.get_index_statistic(
        DifferingVisibilityCountStatistic.STATS_TYPE,
        TEST_INDEX,
        TEST_TAG)
    assert isinstance(differing_visibility_stat, DifferingVisibilityCountStatistic)
    assert differing_visibility_stat.get_tag() == TEST_TAG
    assert differing_visibility_stat.get_index_name() == TEST_INDEX
    assert differing_visibility_stat.get_description() is not None
    assert differing_visibility_stat.get_binning_strategy() is None
    assert isinstance(differing_visibility_stat.get_statistic_type(), IndexStatisticType)
    assert differing_visibility_stat.get_statistic_type().get_string() == 'DIFFERING_VISIBILITY_COUNT'
    assert test_ds.get_statistic_value(differing_visibility_stat) == 0

    # test alternate constructors
    differing_visibility_stat = DifferingVisibilityCountStatistic(TEST_INDEX)
    differing_visibility_stat.set_internal()
    assert differing_visibility_stat.get_index_name() == TEST_INDEX
    assert differing_visibility_stat.get_tag() == INTERNAL_TAG


def test_duplicate_entry_count_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    duplicate_entry_stat = DuplicateEntryCountStatistic()
    duplicate_entry_stat.set_tag(TEST_TAG)
    duplicate_entry_stat.set_index_name(TEST_INDEX)
    test_ds.add_statistic(duplicate_entry_stat)
    write_test_data(test_ds, index)

    # then
    duplicate_entry_stat = test_ds.get_index_statistic(
        DuplicateEntryCountStatistic.STATS_TYPE,
        TEST_INDEX,
        TEST_TAG)
    assert isinstance(duplicate_entry_stat, DuplicateEntryCountStatistic)
    assert duplicate_entry_stat.get_tag() == TEST_TAG
    assert duplicate_entry_stat.get_index_name() == TEST_INDEX
    assert duplicate_entry_stat.get_description() is not None
    assert duplicate_entry_stat.get_binning_strategy() is None
    assert isinstance(duplicate_entry_stat.get_statistic_type(), IndexStatisticType)
    assert duplicate_entry_stat.get_statistic_type().get_string() == 'DUPLICATE_ENTRY_COUNT'
    assert test_ds.get_statistic_value(duplicate_entry_stat) == 0

    # test alternate constructors
    duplicate_entry_stat = DuplicateEntryCountStatistic(TEST_INDEX)
    duplicate_entry_stat.set_internal()
    assert duplicate_entry_stat.get_index_name() == TEST_INDEX
    assert duplicate_entry_stat.get_tag() == INTERNAL_TAG


def test_field_visibility_count_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    field_visibility_stat = FieldVisibilityCountStatistic()
    field_visibility_stat.set_tag(TEST_TAG)
    field_visibility_stat.set_index_name(TEST_INDEX)
    test_ds.add_statistic(field_visibility_stat)
    write_test_data(test_ds, index)

    # then
    field_visibility_stat = test_ds.get_index_statistic(
        FieldVisibilityCountStatistic.STATS_TYPE,
        TEST_INDEX,
        TEST_TAG)
    assert isinstance(field_visibility_stat, FieldVisibilityCountStatistic)
    assert field_visibility_stat.get_tag() == TEST_TAG
    assert field_visibility_stat.get_index_name() == TEST_INDEX
    assert field_visibility_stat.get_description() is not None
    assert field_visibility_stat.get_binning_strategy() is None
    assert isinstance(field_visibility_stat.get_statistic_type(), IndexStatisticType)
    assert field_visibility_stat.get_statistic_type().get_string() == 'FIELD_VISIBILITY_COUNT'
    visibility_counts = test_ds.get_statistic_value(field_visibility_stat)
    assert isinstance(visibility_counts, dict)
    assert len(visibility_counts) == 1
    assert b'' in visibility_counts
    assert visibility_counts[b''] == 360

    # test alternate constructors
    field_visibility_stat = FieldVisibilityCountStatistic(TEST_INDEX)
    field_visibility_stat.set_internal()
    assert field_visibility_stat.get_index_name() == TEST_INDEX
    assert field_visibility_stat.get_tag() == INTERNAL_TAG


def test_index_meta_data_set_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    index_meta_data_set_stat = IndexMetaDataSetStatistic()
    index_meta_data_set_stat.set_tag(TEST_TAG)
    index_meta_data_set_stat.set_index_name(TEST_INDEX)
    test_ds.add_statistic(index_meta_data_set_stat)
    write_test_data(test_ds, index)

    # then
    index_meta_data_set_stat = test_ds.get_index_statistic(
        IndexMetaDataSetStatistic.STATS_TYPE,
        TEST_INDEX,
        TEST_TAG)
    assert isinstance(index_meta_data_set_stat, IndexMetaDataSetStatistic)
    assert index_meta_data_set_stat.get_tag() == TEST_TAG
    assert index_meta_data_set_stat.get_index_name() == TEST_INDEX
    assert index_meta_data_set_stat.get_description() is not None
    assert index_meta_data_set_stat.get_binning_strategy() is None
    assert isinstance(index_meta_data_set_stat.get_statistic_type(), IndexStatisticType)
    assert index_meta_data_set_stat.get_statistic_type().get_string() == 'INDEX_METADATA'
    assert len(test_ds.get_statistic_value(index_meta_data_set_stat)) == 0

    # test alternate constructors
    index_meta_data_set_stat = IndexMetaDataSetStatistic(TEST_INDEX)
    index_meta_data_set_stat.set_internal()
    assert index_meta_data_set_stat.get_index_name() == TEST_INDEX
    assert index_meta_data_set_stat.get_tag() == INTERNAL_TAG


def test_max_duplicates_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    max_duplicates_stat = MaxDuplicatesStatistic()
    max_duplicates_stat.set_tag(TEST_TAG)
    max_duplicates_stat.set_index_name(TEST_INDEX)
    test_ds.add_statistic(max_duplicates_stat)
    write_test_data(test_ds, index)

    # then
    max_duplicates_stat = test_ds.get_index_statistic(
        MaxDuplicatesStatistic.STATS_TYPE,
        TEST_INDEX,
        TEST_TAG)
    assert isinstance(max_duplicates_stat, MaxDuplicatesStatistic)
    assert max_duplicates_stat.get_tag() == TEST_TAG
    assert max_duplicates_stat.get_index_name() == TEST_INDEX
    assert max_duplicates_stat.get_description() is not None
    assert max_duplicates_stat.get_binning_strategy() is None
    assert isinstance(max_duplicates_stat.get_statistic_type(), IndexStatisticType)
    assert max_duplicates_stat.get_statistic_type().get_string() == 'MAX_DUPLICATES'
    assert test_ds.get_statistic_value(max_duplicates_stat) == 0

    # test alternate constructors
    max_duplicates_stat = MaxDuplicatesStatistic(TEST_INDEX)
    max_duplicates_stat.set_internal()
    assert max_duplicates_stat.get_index_name() == TEST_INDEX
    assert max_duplicates_stat.get_tag() == INTERNAL_TAG


def test_partitions_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    partitions_stat = PartitionsStatistic()
    partitions_stat.set_tag(TEST_TAG)
    partitions_stat.set_index_name(TEST_INDEX)
    test_ds.add_statistic(partitions_stat)
    write_test_data(test_ds, index)

    # then
    partitions_stat = test_ds.get_index_statistic(
        PartitionsStatistic.STATS_TYPE,
        TEST_INDEX,
        TEST_TAG)
    assert isinstance(partitions_stat, PartitionsStatistic)
    assert partitions_stat.get_tag() == TEST_TAG
    assert partitions_stat.get_index_name() == TEST_INDEX
    assert partitions_stat.get_description() is not None
    assert partitions_stat.get_binning_strategy() is None
    assert isinstance(partitions_stat.get_statistic_type(), IndexStatisticType)
    assert partitions_stat.get_statistic_type().get_string() == 'PARTITIONS'
    partitions = test_ds.get_statistic_value(partitions_stat)
    assert isinstance(partitions, set)
    assert len(partitions) >= 1
    for item in partitions:
        assert isinstance(item, bytes)

    # test alternate constructors
    partitions_stat = PartitionsStatistic(TEST_INDEX)
    partitions_stat.set_internal()
    assert partitions_stat.get_index_name() == TEST_INDEX
    assert partitions_stat.get_tag() == INTERNAL_TAG


def test_row_range_histogram_statistic(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    row_range_histogram_stat = RowRangeHistogramStatistic()
    row_range_histogram_stat.set_tag(TEST_TAG)
    row_range_histogram_stat.set_index_name(TEST_INDEX)
    test_ds.add_statistic(row_range_histogram_stat)
    write_test_data(test_ds, index)

    # then
    row_range_histogram_stat = test_ds.get_index_statistic(
        RowRangeHistogramStatistic.STATS_TYPE,
        TEST_INDEX,
        TEST_TAG)
    assert isinstance(row_range_histogram_stat, RowRangeHistogramStatistic)
    assert row_range_histogram_stat.get_tag() == TEST_TAG
    assert row_range_histogram_stat.get_index_name() == TEST_INDEX
    assert row_range_histogram_stat.get_description() is not None
    assert row_range_histogram_stat.get_binning_strategy() is None
    assert isinstance(row_range_histogram_stat.get_statistic_type(), IndexStatisticType)
    assert row_range_histogram_stat.get_statistic_type().get_string() == 'ROW_RANGE_HISTOGRAM'
    histogram = test_ds.get_statistic_value(row_range_histogram_stat)
    assert isinstance(histogram, NumericHistogram)
    assert histogram.get_min_value() == float('inf')
    assert histogram.get_max_value() == float('-inf')
    assert histogram.sum(histogram.quantile(1.0)) == 360
    assert histogram.get_total_count() == 360

    # test alternate constructors
    row_range_histogram_stat = RowRangeHistogramStatistic(TEST_INDEX)
    row_range_histogram_stat.set_internal()
    assert row_range_histogram_stat.get_index_name() == TEST_INDEX
    assert row_range_histogram_stat.get_tag() == INTERNAL_TAG


def test_composite_binning_strategy(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    binning_strategy = CompositeBinningStrategy([
        FieldValueBinningStrategy([POINT_COLOR_FIELD]),
        FieldValueBinningStrategy([POINT_SHAPE_FIELD]),
    ])
    count_stat = CountStatistic()
    count_stat.set_tag(TEST_TAG)
    count_stat.set_type_name(POINT_TYPE_NAME)
    count_stat.set_binning_strategy(binning_strategy)
    test_ds.add_statistic(count_stat)
    write_test_data(test_ds, index)

    # then
    count_stat = test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, TEST_TAG)
    binning_strategy = count_stat.get_binning_strategy()
    assert isinstance(binning_strategy, CompositeBinningStrategy)
    assert binning_strategy.get_strategy_name() == 'COMPOSITE'
    assert binning_strategy.get_description() is not None
    assert test_ds.get_statistic_value(count_stat) == 360
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat))
    # There should be one bin for every color/shape combination
    assert len(binned_values) == 12
    for b, v in binned_values:
        assert isinstance(b, bytes)
        # colors and shapes are evenly distributed, so each bin should have 30
        assert v == 30


def test_data_type_binning_strategy(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    binning_strategy = DataTypeBinningStrategy()
    count_stat = CountStatistic()
    count_stat.set_tag(TEST_TAG)
    count_stat.set_type_name(POINT_TYPE_NAME)
    count_stat.set_binning_strategy(binning_strategy)
    test_ds.add_statistic(count_stat)
    write_test_data(test_ds, index)

    # then
    count_stat = test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, TEST_TAG)
    binning_strategy = count_stat.get_binning_strategy()
    assert isinstance(binning_strategy, DataTypeBinningStrategy)
    assert binning_strategy.get_strategy_name() == 'DATA_TYPE'
    assert binning_strategy.get_description() is not None
    assert test_ds.get_statistic_value(count_stat) == 360
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat))
    # The only bin is for our point adapter
    assert len(binned_values) == 1
    assert binning_strategy.bin_to_string(binned_values[0][0]) == POINT_TYPE_NAME
    assert binned_values[0][1] == 360

    # test bin constraint
    bin_constraint = BinConstraints.of(POINT_TYPE_NAME.encode())
    assert isinstance(bin_constraint, BinConstraints)
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat, bin_constraint))
    # There should be one bin for every 5 values
    assert len(binned_values) == 1
    assert binning_strategy.bin_to_string(binned_values[0][0]) == POINT_TYPE_NAME
    assert binned_values[0][1] == 360


def test_field_value_binning_strategy(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    binning_strategy = FieldValueBinningStrategy([POINT_COLOR_FIELD, POINT_SHAPE_FIELD])
    count_stat = CountStatistic()
    count_stat.set_tag(TEST_TAG)
    count_stat.set_type_name(POINT_TYPE_NAME)
    count_stat.set_binning_strategy(binning_strategy)
    test_ds.add_statistic(count_stat)
    write_test_data(test_ds, index)

    # then
    count_stat = test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, TEST_TAG)
    binning_strategy = count_stat.get_binning_strategy()
    assert isinstance(binning_strategy, FieldValueBinningStrategy)
    assert binning_strategy.get_strategy_name() == 'FIELD_VALUE'
    assert binning_strategy.get_description() is not None
    assert test_ds.get_statistic_value(count_stat) == 360
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat))
    # There should be one bin for every color/shape combination
    assert len(binned_values) == 12
    for b, v in binned_values:
        assert isinstance(b, bytes)
        # colors and shapes are evenly distributed, so each bin should have 30
        assert v == 30

    # test bin prefix constraint
    bin_constraint = BinConstraints.of_prefix(b'BLUE')
    assert isinstance(bin_constraint, BinConstraints)
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat, bin_constraint))
    # There should be 4 bins, one for each shape
    assert len(binned_values) == 4
    for b, v in binned_values:
        assert isinstance(b, bytes)
        # each bin should have 30
        assert v == 30


def test_numeric_range_field_value_binning_strategy(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    binning_strategy = NumericRangeFieldValueBinningStrategy(fields=[POINT_NUMBER_FIELD], interval=5)
    count_stat = CountStatistic()
    count_stat.set_tag(TEST_TAG)
    count_stat.set_type_name(POINT_TYPE_NAME)
    count_stat.set_binning_strategy(binning_strategy)
    test_ds.add_statistic(count_stat)
    write_test_data(test_ds, index)

    # then
    count_stat = test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, TEST_TAG)
    binning_strategy = count_stat.get_binning_strategy()
    assert isinstance(binning_strategy, NumericRangeFieldValueBinningStrategy)
    assert binning_strategy.get_strategy_name() == 'NUMERIC_RANGE'
    assert binning_strategy.get_description() is not None
    assert test_ds.get_statistic_value(count_stat) == 360
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat))
    # There should be one bin for every 5 values
    assert len(binned_values) == 72
    for b, v in binned_values:
        assert isinstance(b, bytes)
        assert v == 5

    # test numeric range constraint
    bin_constraint = BinConstraints.of_object(Range(0, 180))
    assert isinstance(bin_constraint, BinConstraints)
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat, bin_constraint))
    # There should be one bin for every 5 values
    assert len(binned_values) == 36
    for b, v in binned_values:
        assert isinstance(b, bytes)
        assert v == 5


def test_time_range_field_value_binning_strategy(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    binning_strategy = TimeRangeFieldValueBinningStrategy(fields=[POINT_TIME_FIELD], periodicity='minute')
    count_stat = CountStatistic()
    count_stat.set_tag(TEST_TAG)
    count_stat.set_type_name(POINT_TYPE_NAME)
    count_stat.set_binning_strategy(binning_strategy)
    test_ds.add_statistic(count_stat)
    write_test_data(test_ds, index)

    # then
    count_stat = test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, TEST_TAG)
    binning_strategy = count_stat.get_binning_strategy()
    assert isinstance(binning_strategy, TimeRangeFieldValueBinningStrategy)
    assert binning_strategy.get_strategy_name() == 'TIME_RANGE'
    assert binning_strategy.get_description() is not None
    assert test_ds.get_statistic_value(count_stat) == 360
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat))
    # Each value is 1 second apart, so there should be 6 bins with a periodicity of minute
    assert len(binned_values) == 6
    for b, v in binned_values:
        assert isinstance(b, bytes)
        assert v == 60

    # test time range constraint
    bin_constraint = BinConstraints.of_object(Interval(datetime.utcfromtimestamp(0), datetime.utcfromtimestamp(180)))
    assert isinstance(bin_constraint, BinConstraints)
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat, bin_constraint))
    # Each value is 1 second apart, so there should be 3 bins with a periodicity of minute
    assert len(binned_values) == 3
    for b, v in binned_values:
        assert isinstance(b, bytes)
        assert v == 60


def test_spatial_field_value_binning_strategy(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    binning_strategy = SpatialFieldValueBinningStrategy(fields=[POINT_GEOMETRY_FIELD], type='GEOHASH', precision=1)
    count_stat = CountStatistic()
    count_stat.set_tag(TEST_TAG)
    count_stat.set_type_name(POINT_TYPE_NAME)
    count_stat.set_binning_strategy(binning_strategy)
    test_ds.add_statistic(count_stat)
    write_test_data_offset(test_ds, index)

    # then
    count_stat = test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, TEST_TAG)
    binning_strategy = count_stat.get_binning_strategy()
    assert isinstance(binning_strategy, SpatialFieldValueBinningStrategy)
    assert binning_strategy.get_strategy_name() == 'SPATIAL'
    assert binning_strategy.get_description() is not None
    assert test_ds.get_statistic_value(count_stat) == 360
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat))
    assert len(binned_values) == 8
    for b, v in binned_values:
        assert isinstance(b, bytes)
        assert v == 45

    # test polygon constraint
    bin_constraint = BinConstraints.of_object(Polygon([[0.5, 0.5], [0.5, 45.5], [45.5, 45.5], [45.5, 0.5], [0.5, 0.5]]))
    assert isinstance(bin_constraint, BinConstraints)
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat, bin_constraint))
    # There is a bin every 45 degrees so there should be 2 bins
    assert len(binned_values) == 2
    for b, v in binned_values:
        assert isinstance(b, bytes)
        assert v == 45

    bin_constraint = BinConstraints.of_object(Envelope(min_x=1, min_y=1, max_x=91, max_y=90))
    assert isinstance(bin_constraint, BinConstraints)
    assert test_ds.get_statistic_value(count_stat, bin_constraint) == 135



def test_partition_binning_strategy(test_ds):
    # given
    index = SpatialIndexBuilder().set_name(TEST_INDEX).create_index()
    adapter = POINT_TYPE_ADAPTER
    test_ds.add_type(adapter, index)

    # when
    binning_strategy = PartitionBinningStrategy()
    count_stat = CountStatistic()
    count_stat.set_tag(TEST_TAG)
    count_stat.set_type_name(POINT_TYPE_NAME)
    count_stat.set_binning_strategy(binning_strategy)
    test_ds.add_statistic(count_stat)
    write_test_data(test_ds, index)

    # then
    count_stat = test_ds.get_data_type_statistic(CountStatistic.STATS_TYPE, POINT_TYPE_NAME, TEST_TAG)
    binning_strategy = count_stat.get_binning_strategy()
    assert isinstance(binning_strategy, PartitionBinningStrategy)
    assert binning_strategy.get_strategy_name() == 'PARTITION'
    assert binning_strategy.get_description() is not None
    assert test_ds.get_statistic_value(count_stat) == 360
    binned_values = results_as_list(test_ds.get_binned_statistic_values(count_stat))
    # No real assumptions can be made about how the data is partitioned, but there will be at least 1 and all values
    # should sum to 360
    assert len(binned_values) >= 1
    assert reduce(lambda a, b: a + b, map(lambda a: a[1], binned_values)) == 360



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base.type_conversions import StringArrayType
from pygw.config import geowave_pkg
from .field_value_binning_strategy import FieldValueBinningStrategy


class TimeRangeFieldValueBinningStrategy(FieldValueBinningStrategy):
    """
    Statistic binning strategy that bins statistic values by the temporal representation of the value of a given field.
    It bins time values by a temporal periodicity (any time unit), default to daily. A statistic using this binning
    strategy can be constrained using an Interval as a constraint).
    """

    def __init__(self, fields=None, time_zone="GMT", periodicity="day", java_ref=None):
        if java_ref is None:
            j_periodicity = geowave_pkg.core.geotime.index.dimension.TemporalBinningStrategy.Unit.fromString(
                periodicity)
            java_ref = geowave_pkg.core.geotime.store.statistics.binning.TimeRangeFieldValueBinningStrategy(
                time_zone, j_periodicity, StringArrayType().to_java(fields))
        super().__init__(fields, java_ref)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from .geowave_object import GeoWaveObject
from .java_transformer import JavaTransformer
from ..config import java_pkg


class Range(GeoWaveObject):
    def __init__(self, minimum=0, maximum=0, java_ref=None):
        if java_ref is None:
            java_ref = java_pkg.org.apache.commons.lang3.Range.between(float(minimum), float(maximum))
        super().__init__(java_ref)

    def get_minimum(self):
        return self._java_ref.getMinimum()

    def get_maximum(self):
        return self._java_ref.getMaximum()


class RangeTransformer(JavaTransformer):
    """
    Transforms Java Ranges into a pygw Range.
    """

    def transform(self, j_object):
        """
        Transform the given Java Range into a tuple.

        Args:
            j_object (Java Range): A Range Java object.
        Returns:
            A pygw implementation of Range.
        """
        return Range(java_ref=j_object)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
from pygw.base.type_conversions import StringArrayType
from pygw.config import geowave_pkg
from ..statistic_binning_strategy import StatisticBinningStrategy


class FieldValueBinningStrategy(StatisticBinningStrategy):
    """
    Statistic binning strategy that bins statistic values by the string representation of the value of one or more
    fields.
    """

    def __init__(self, fields=None, java_ref=None):
        if java_ref is None:
            java_ref = geowave_pkg.core.store.statistics.binning.FieldValueBinningStrategy(
                StringArrayType().to_java(fields))
        super().__init__(java_ref)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.base import GeoWaveObject

from ..base.java_transformer import JavaTransformer


class Query(GeoWaveObject):
    """
    Base Query class that wraps a GeoWave query.
    """

    def __init__(self, java_ref, java_transformer):
        if java_transformer is not None:
            assert isinstance(java_transformer, JavaTransformer)
        self.java_transformer = java_transformer
        super().__init__(java_ref)



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from pygw.base import GeoWaveObject


class SimpleFeatureType(GeoWaveObject):
    """
    Defines a schema for vector features.
    """

    def __init__(self, java_ref, attributes):
        self.attribute_dict = {}
        for a in attributes:
            self.attribute_dict[a.descriptor] = a
        self.attribute_list = attributes
        super().__init__(java_ref)

    def get_type_name(self):
        """
        Returns:
            The name of the feature type.
        """
        return self._java_ref.getTypeName()

    def get_attribute(self, attribute):
        """
        Gets an attribute descriptor by index or by name.

        Args:
            attribute (str or int): Name or index of the descriptor to get.
        Returns:
            A `pygw.geotools.attribute_descriptor.AttributeDescriptor` for the attribute requested.
        """
        if isinstance(attribute, int):
            return self.attribute_list[attribute]
        elif isinstance(attribute, str) and attribute in self.attribute_dict:
            return self.attribute_dict[attribute]
        return None

    def get_attribute_descriptors(self):
        """
        Gets all of the attribute descriptors for this feature type.

        Returns:
            A list of `pygw.geotools.attribute_descriptor.AttributeDescriptor` for this feature type.
        """
        return self.attribute_list



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from .geowave_object import GeoWaveObject


class DataTypeAdapter(GeoWaveObject):
    """
    Base class for data type adapters.
    """

    def get_type_name(self):
        """
        Returns:
            The type name of the data adapter.
        """
        return self._java_ref.getTypeName()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================
"""
This module contains classes that are used for GeoWave statistics.

It contains the following import shortcuts:
```python
from pygw.statistics import Statistic
from pygw.statistics import DataTypeStatistic
from pygw.statistics import FieldStatistic
from pygw.statistics import IndexStatistic
from pygw.statistics import StatisticValue
from pygw.statistics import StatisticBinningStrategy
from pygw.statistics import StatisticType
from pygw.statistics import DataTypeStatisticType
from pygw.statistics import FieldStatisticType
from pygw.statistics import IndexStatisticType
from pygw.statistics import BinConstraints
```
"""

from .statistic import Statistic
from .statistic import DataTypeStatistic
from .statistic import FieldStatistic
from .statistic import IndexStatistic
from .statistic_value import StatisticValue
from .statistic_binning_strategy import StatisticBinningStrategy
from .statistic_type import StatisticType
from .statistic_type import DataTypeStatisticType
from .statistic_type import FieldStatisticType
from .statistic_type import IndexStatisticType
from .bin_constraints import BinConstraints




#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

from .geowave_object import GeoWaveObject


class WriteResults(GeoWaveObject):
    """
    Contains the results of a write operation.
    """

    def __init__(self, java_ref):
        super().__init__(java_ref)

    def get_written_index_names(self):
        """
        Gets a list of all the index names that were written to during the write.

        Returns:
            A list of index names that were written to.
        """
        j_index_names = self._java_ref.getWrittenIndexNames().iterator()
        values = []
        while j_index_names.hasNext():
            values.append(j_index_names.next())
        return values

    def get_insertion_ids_written(self, index_name):
        """
        Gets the insertion IDs that were written to the index with the given index name.

        Args:
            index_name (str): The name of the index.
        Returns:
            A list of insertion ids that were written into the index.
        """
        j_insertion_ids = self._java_ref.getInsertionIdsWritten(index_name).getCompositeInsertionIds().iterator()
        values = []
        while j_insertion_ids.hasNext():
            values.append(j_insertion_ids.next())
        return values

    def is_empty(self):
        """
        Returns:
            True if nothing was written, False otherwise.
        """
        return self._java_ref.isEmpty()



#
# Copyright (c) 2013-2022 Contributors to the Eclipse Foundation

#
# See the NOTICE file distributed with this work for additional information regarding copyright
# ownership. All rights reserved. This program and the accompanying materials are made available
# under the terms of the Apache License, Version 2.0 which accompanies this distribution and is
# available at http://www.apache.org/licenses/LICENSE-2.0.txt
# ===============================================================================================

import pytest
import os
import shutil
import time

from datetime import datetime

from shapely.geometry import Point

from pygw.base import CloseableIterator
from pygw.store import DataStoreFactory
from pygw.store.rocksdb import RocksDBOptions
from pygw.geotools import SimpleFeatureTypeBuilder
from pygw.geotools import AttributeDescriptor
from pygw.geotools import FeatureDataAdapter
from pygw.geotools import SimpleFeatureBuilder

# "Point" Type
POINT_TYPE_NAME = "TestPointType"
POINT_GEOMETRY_FIELD = "the_geom"
POINT_TIME_FIELD = "date"
POINT_NUMBER_FIELD = "flt"
POINT_COLOR_FIELD = "color"
POINT_SHAPE_FIELD = "shape"
_point_type_builder = SimpleFeatureTypeBuilder()
_point_type_builder.set_name(POINT_TYPE_NAME)
_point_type_builder.add(AttributeDescriptor.point(POINT_GEOMETRY_FIELD))
_point_type_builder.add(AttributeDescriptor.date(POINT_TIME_FIELD))
_point_type_builder.add(AttributeDescriptor.float(POINT_NUMBER_FIELD))
_point_type_builder.add(AttributeDescriptor.string(POINT_COLOR_FIELD))
_point_type_builder.add(AttributeDescriptor.string(POINT_SHAPE_FIELD))
POINT_TYPE = _point_type_builder.build_feature_type()

# "Point" Type Adapter
POINT_TYPE_ADAPTER = FeatureDataAdapter(POINT_TYPE)

# "Point" Feature builder
POINT_FEATURE_BUILDER = SimpleFeatureBuilder(POINT_TYPE)

COLORS = ['RED', 'GREEN', 'BLUE']
SHAPES = ['SQUARE', 'CIRCLE', 'TRIANGLE', 'RECTANGLE']


def _create_feature(fid, geometry, timestamp):
    POINT_FEATURE_BUILDER.set_attr(POINT_GEOMETRY_FIELD, geometry)
    POINT_FEATURE_BUILDER.set_attr(POINT_TIME_FIELD, datetime.utcfromtimestamp(timestamp))
    POINT_FEATURE_BUILDER.set_attr(POINT_NUMBER_FIELD, timestamp)
    POINT_FEATURE_BUILDER.set_attr(POINT_COLOR_FIELD, COLORS[timestamp % 3])
    POINT_FEATURE_BUILDER.set_attr(POINT_SHAPE_FIELD, SHAPES[timestamp % 4])
    return POINT_FEATURE_BUILDER.build(fid)


def latitude(lon_value):
    if lon_value < 0:
        return lon_value % -90
    return lon_value % 90


TEST_DATA = [
    _create_feature(id_, Point(i, latitude(i)), i) for
    id_, i in enumerate(range(-180, 180))]

TEST_DATA_OFFSET = [
    _create_feature(id_, Point(i+0.5, latitude(i+0.5)), i) for
    id_, i in enumerate(range(-180, 180))]

# Test Directory
TEST_DIR = os.path.join(os.getcwd(), "test")


@pytest.fixture
def test_ds():
    os.makedirs(TEST_DIR, exist_ok=True)
    options = RocksDBOptions()
    options.set_geowave_namespace("geowave.tests")
    options.set_directory(os.path.join(TEST_DIR, "datastore"))
    ds = DataStoreFactory.create_data_store(options)
    yield ds
    # teardown here
    ds.delete_all()
    shutil.rmtree(TEST_DIR)
    while os.path.isdir(TEST_DIR):
        time.sleep(0.01)

def write_test_data_offset(ds, *expected_indices):
    write_test_data(ds, *expected_indices, data=TEST_DATA_OFFSET)

def write_test_data(ds, *expected_indices, data=TEST_DATA):
    writer = ds.create_writer(POINT_TYPE_ADAPTER.get_type_name())
    for pt in data:
        results = writer.write(pt)
        assert not results.is_empty()
        written_indices = results.get_written_index_names()
        assert len(written_indices) == len(expected_indices)
        assert all([idx.get_name() in written_indices for idx in expected_indices])
    writer.close()


def results_as_list(results):
    assert isinstance(results, CloseableIterator)
    res = [d for d in results]
    results.close()
    return res
